from xai_components.base import InArg, OutArg, InCompArg, Component, BaseComponent, xai_component

@xai_component
class BranchComponent(Component):
    when_true: BaseComponent
    when_false: BaseComponent

    condition: InArg[bool]
    
    def do(self, ctx) -> BaseComponent:
        if self.condition.value:
            next = self.when_true
        else:
            next = self.when_false
        while next:
            next = next.do(ctx)
        try:
            return self.next
        except:
            return None
    
@xai_component
class LoopComponent(Component):
    body: BaseComponent

    condition: InArg[bool]
    
    def do(self, ctx) -> BaseComponent:
        while self.condition.value:
            next_body = self.body.do(ctx)
            while next_body:
                next_body = next_body.do(ctx)
            return self
        return self.next
    
@xai_component
class CounterComponent(Component):
    start_number: InArg[int]
    step: InArg[int]
    out_number: OutArg[int]
    
    state: any
    
    def __init__(self):
        super().__init__()
        self.state = None
        
    def execute(self, ctx) -> None:
        if self.state is None:
            self.state = self.start_number.value
        else:
            self.state += self.step.value
            
        self.out_number.value = self.state

@xai_component
class ComparisonComponent(Component):
    a: InArg[int]
    b: InArg[int]
    op: InArg[str]
    
    out: OutArg[bool]

    def execute(self, ctx) -> None:
        self.out.value = eval(str(self.a.value) + " " + self.op.value + " " + str(self.b.value))


class MutableVariable:
    _fn: any

    def __init__(self):
        self._fn = None
    
    def set_fn(self, fn) -> None:
        self._fn = fn
        
    @property
    def value(self) -> any:
        return self._fn()
    
    
@xai_component
class GetVariableComponent(Component):
    name: InArg[str]
    value: OutArg[any]
    
    def __init__(self):
        super().__init__()
        self.value = MutableVariable()
        
    def execute(self, ctx) -> None:
        self.value.set_fn(lambda: ctx[self.name.value])

@xai_component
class SetVariableComponent(Component):
    name: InArg[str]
    value: InArg[any]

    def execute(self, ctx) -> None:
        ctx[self.name.value] = self.value.value


@xai_component
class DefineVariableComponent(Component):
    name: InArg[str]
    value: InArg[any]
    ref: OutArg[any]

    
    def __init__(self):
        super().__init__()
        self.ref = MutableVariable()
        
    def execute(self, ctx) -> None:
        ctx[self.name.value] = self.value.value
        self.ref.set_fn(lambda: ctx[self.name.value])



import os
from git import Repo
from git.remote import RemoteProgress
from pathlib import Path

class Progress(RemoteProgress):
    def update(self, *args):
        print(self._cur_line, end='\r', flush=True)

def get_submodule_config(user_query):
    
    import configparser
    config = configparser.ConfigParser()
    
    config.read('.xircuits/.gitmodules')
    
    submodule_keys = [submodule for submodule in config.sections() if user_query in submodule]
    if len(submodule_keys) == 0:
        print(user_query + " component library submodule not found.")
        return
        
    if len(submodule_keys) > 1:
        raise ValueError(f"Multiple instances of '{user_query}' found.")

    submodule_key = submodule_keys.pop(0)
    
    submodule_path = config[submodule_key]["path"]
    submodule_url = config[submodule_key]["url"]
    
    return submodule_path, submodule_url


def request_submodule_library(component_library_query):
    
    submodule_path, submodule_url = get_submodule_config(component_library_query)
    
    print("Cloning " + submodule_path + " from " + submodule_url)
    Repo.clone_from(submodule_url, submodule_path, progress=Progress())


def get_submodules(repo, ref="master"):
    try:
        gitmodules_content = repo.get_contents(".gitmodules", ref=ref)
        gitmodules = gitmodules_content.decoded_content.decode("utf-8")
        
        submodules = []
        for line in gitmodules.split("\n"):
            if "path = " in line:
                submodules.append(line.split(" = ")[-1].strip())
        return submodules
    
    except:
        return []


from typing import List
from dataclasses import dataclass

from xircuits.compiler.port import Port


@dataclass
class Node:
    id: str
    name: str
    type: str
    file: str
    ports: List[Port]

    def generate_import(self):
        pass

    def generate_instantiation(self):
        pass

    def generate_assignments(self):
        pass

    def __hash__(self):
        return id.__hash__()


from pyspark.sql import SparkSession
from pyspark.sql import Row

from datetime import datetime, date
import matplotlib.pyplot as plt

from xai_components.base import InArg, InCompArg, OutArg, Component, xai_component
import json
import os
import sys


os.environ['PYSPARK_PYTHON'] = sys.executable
os.environ['PYSPARK_DRIVER_PYTHON'] = sys.executable

@xai_component
class xSparkSession(Component):
    """Instantiates a Spark Session.
  
    ##### inPorts:
    - master: Cluster URL to connect to (e.g. mesos://host:port, spark://host:port, local[4]). 
        Default: `local`.
    - appname: A name for your job, to display on the cluster web UI.
        Default: .xircuits canvas name.
    - config: Runtime configuration interface for Spark.
        Default: `" "`.

    ##### outPorts:
    - sparksession: A spark session instance.
    """

    master: InArg[str]  
    appname: InArg[str]
    config: InArg[str]
    sparksession: OutArg[any]

    def execute(self, ctx) -> None:

        master_url = self.master.value if self.master.value else "local"
        app_name = self.appname.value if self.appname.value else os.path.splitext(sys.argv[0])[0]
        config = self.config.value if self.config.value else " "

        if self.master.value or self.config.value:
            spark = SparkSession.builder.master(master_url) \
                                .appName(app_name) \
                                .config(config) \
                                .getOrCreate()

        else:
            spark = SparkSession.builder.getOrCreate()
        
        self.sparksession.value = spark

@xai_component
class SparkReadPandas(Component):
    """Creates a Spark df from a pandas df.
  
    ##### inPorts:
    - in_sparksession: A spark session.
    - pandas_dataframe: pandas dataframe.

    ##### outPorts:
    - out_sparksession: A spark session.
    - out_dataframe: A spark dataframe.
    """

    in_sparksession: InCompArg[any]
    pandas_dataframe: InCompArg[any]
    out_sparksession: OutArg[any]
    out_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        df = self.pandas_dataframe.value
        spark_df = spark.createDataFrame(df)
        spark_df.show()

        self.out_sparksession.value = spark
        self.out_dataframe.value = spark_df

@xai_component
class SparkReadFile(Component):
    """Reads a Spark supported file format (json / csv / parquet / orc ) 
    and outputs a Spark dataframe.

    ##### inPorts:
    - in_sparksession: A spark session.
    - file_input: a Spark supported file format (json / csv / parquet / orc ) filepath.

    ##### outPorts:
    - out_sparksession: A spark session.
    - out_dataframe: A spark dataframe.
    """
    in_sparksession: InCompArg[any]
    file_input: InCompArg[str]
    out_sparksession: OutArg[any]
    out_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        filepath = self.file_input.value
        ext = filepath.split(".")[-1]

        if ext == "csv":
            df = spark.read.load(filepath,
                     format="csv", sep=",", inferSchema="true", header="true")
            df.show()

        elif ext == "parquet":
            df = spark.read.load(filepath)
            df.show()

        elif ext == "orc":
            df = spark.read.orc(filepath)
            df.show()

        elif ext == "json":
            df = spark.read.load("filepath", format="json")

        else:
            print("Unrecognized file format! Please input json / csv / parquet / orc ")

        self.out_sparksession.value = spark
        self.out_dataframe.value = df

@xai_component
class SparkReadCSV(Component):
    """Reads a csv and outputs a Spark dataframe.

    ##### inPorts:
    - in_sparksession: a spark session.
    - file_input: a csv filepath.
    - separator: the data separator in csv. Default `,`.
    - header: bool whether csv has headers. Default `True`.

    ##### outPorts:
    - out_sparksession: A spark session.
    - out_dataframe: A spark dataframe.
    """
    in_sparksession: InCompArg[any]
    file_input: InCompArg[str]
    separator: InArg[str]
    header: InArg[bool]
    out_sparksession: OutArg[any]
    out_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        filepath = self.file_input.value

        sep = self.separator.value if self.separator.value else ","
        header = self.header.value
        
        df = spark.read.load(filepath,
                 format="csv", sep=sep, inferSchema="true", header=header)

        df.show()

        self.out_sparksession.value = spark
        self.out_dataframe.value = df

@xai_component
class SparkWriteFile(Component):
    """Writes a Spark dataframe to disk in supported output format (csv / parquet / orc).

    ##### inPorts:
    - dataframe: A Spark dataframe.
    - output_name: desired output name. The format will be inferred from the extension.
        Currently supports `csv` / `parquet` / `orc` file format.
    - header: bool whether csv has headers. Default `True`.

    ##### outPorts:
    - out_sparksession: A spark session.
    """
    dataframe: InCompArg[any]
    output_name: InCompArg[str]
    header: InArg[bool]
    out_sparksession: OutArg[any]

    def __init__(self):
        super().__init__()
        self.header.value = True

    def execute(self, ctx) -> None:

        df = self.dataframe.value
        filepath = self.output_name.value
        header = self.header.value
        ext = filepath.split(".")[-1]

        if ext == "csv":
            df.write.csv(filepath, header=header)
        elif ext == "parquet":
            df.write.parquet(filepath)
        elif ext == "orc":
            df.write.orc(filepath)
        else:
            print("Unrecognized file format! Please input csv / parquet / orc.")

@xai_component
class SparkSQL(Component):
    """Performs a SparkSQL query to obtain a Spark dataframe.

    ##### inPorts:
    - in_sparksession: A spark session.
    - dataframe: a Spark dataframe.
    - sql_string: a SQL query to be performed on the dataframe.
    - table_name: specify a table name if not already created by createOrReplaceTempView.

    ##### outPorts:
    - out_sparksession: A spark session.
    - sql_dataframe: the Spark dataframe obtained from the SQL query.
    """
    in_sparksession: InCompArg[any]
    dataframe: InCompArg[any]
    sql_string: InCompArg[str]
    table_name: InArg[str]

    out_sparksession: OutArg[any]
    sql_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        df = self.dataframe.value
        sql_string = self.sql_string.value
        table_name = self.table_name.value if self.table_name.value else ""

        if table_name:
            df.createOrReplaceTempView(table_name)

        else:
            #handler if user does not specify table name
            df.createOrReplaceTempView("tableA")
            sql_string = sql_string + " FROM tableA"
        
        sql_df = spark.sql(sql_string)
        sql_df.show()
        
        self.out_sparksession.value = spark
        self.sql_dataframe.value = sql_df
@xai_component
class SparkVisualize(Component):
    """Visualizes a Spark dataframe.

    ##### inPorts:
    - dataframe: a Spark dataframe.
    - plot_type: the type of plot to be generated. Currently support `bar`, `scatter` and `line` plots. 
        Default `bar` chart.
    - x_axis: the X axis / variable to be visualized.
    - y_axis:the Y axis to be benchmarked on. 
        Default is `None`, which means it will visualize based on raw numbers. 
    - output_name: the chart name to be saved as `visual.png`

    """
    dataframe: InCompArg[any]
    plot_type: InArg[str]
    x_axis: InCompArg[str]
    y_axis: InArg[str]
    output_name: InArg[str]

    def execute(self, ctx) -> None:
        
        df = self.dataframe.value
        plot_type = self.plot_type.value if self.plot_type.value else "bar"
        output_name = self.output_name.value if self.output_name.value else "visual.png"
        x_axis = self.x_axis.value
        y_axis = self.y_axis.value

        pd_df = df.toPandas()

        if plot_type == 'bar':
            pd_df[x_axis].value_counts().plot(kind='bar')

        elif plot_type == 'scatter':
            pd_df.plot(x=x_axis, y=y_axis, kind = 'scatter')    

        elif plot_type == 'line':

            pd_df.plot(x=x_axis, y=y_axis, kind = 'line')    
        
        plt.tight_layout()
        plt.savefig(output_name)
        plt.show()


from xai_components.base import InArg, InCompArg, OutArg, Component, xai_component
import os
import sys
from pyspark.mllib.linalg import SparseVector

os.environ['PYSPARK_PYTHON'] = sys.executable
os.environ['PYSPARK_DRIVER_PYTHON'] = sys.executable


#This should be treated as a parameter component in the future
@xai_component
class SparkSparseVector(Component):
    """Turns a List into a sparse vector.
    ie: sparse_vector = SparseVector(4, [1, 3], [3.0, 4.0])
    
    ### Reference:
    - [Spark Sparse Vector](https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.mllib.linalg.SparseVector.html)

    ##### inPorts:
    - vector_list: List to turn into sparse vector

    ##### outPorts:
    - sparse_vector: the converted sparse vector
    """
    vector_list: InCompArg[list]
    sparse_vector: OutArg[SparseVector]

    def execute(self, ctx) -> None:
        
        from pyspark.mllib.linalg import SparseVector

        vector_list = self.vector_list.value

        sparse_vector = SparseVector(*vector_list)
        sparse_vector

        self.sparse_vector.value = sparse_vector


@xai_component
class SparkLabeledPoint(Component):
    """Creates a labeled point from a dense OR sparse vector.
    A dense vector can be created by supplying a list.
    A sparse vector can be passed from the SparkSparseVector component.
    Only the **sparse vector** input will be used if both are supplied.

    ### Reference:
    - [Spark Labeled Point](https://spark.apache.org/docs/latest/mllib-data-types.html#labeled-point)

    ##### inPorts:
    - label: Label for this data point.
    - dense_vector: can be created by supplying a Literal List.
    - sparse_vector: can be passed from the SparkSparseVector component.

    ##### outPorts:
    - labeled_point: a spark labeled point.
    """

    label: InCompArg[float]
    dense_vector: InArg[list]
    sparse_vector: InArg[SparseVector]

    labeled_point: OutArg[any]

    def execute(self, ctx) -> None:
        
        from pyspark.mllib.regression import LabeledPoint

        label = self.label.value
        vector = self.dense_vector.value if self.dense_vector.value else self.sparse_vector.value

        labeled_point = LabeledPoint(label, vector)
        print(labeled_point)

        self.labeled_point.value = labeled_point


@xai_component
class SparkLoadImageFolder(Component):
    """Loads an image directory as a Spark dataframe.

    ### Reference:
    - [Spark Image Data Source](https://spark.apache.org/docs/latest/ml-datasource.html#image-data-source)

    ##### inPorts:
    - in_sparksession: a spark session.
    - folder_path: path of the image directory

    ##### outPorts:
    - out_sparksession: a spark session.
    - out_dataframe: dataframe that contains the image directory information.
    """
    in_sparksession: InCompArg[any]
    folder_path: InCompArg[str]

    out_sparksession: OutArg[any]
    out_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        folder_path = self.folder_path.value

        df = spark.read.format("image").option("dropInvalid", True).load(folder_path)
        df.show()

        self.out_sparksession.value = spark
        self.out_dataframe.value = df


@xai_component
class SparkSplitDataFrame(Component):
    """Split a spark dataframe into train and test splits.

    ### Reference:
    - [Spark DataFrame Random Split](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.randomSplit.html)

    ##### inPorts:
    - in_dataframe: a Spark DataFrame.
    - train_split: the train split. 
        Default `0.75`.
    - seed: the seed for sampling.

    ##### outPorts:
    - train_dataframe: the DataFrame that contains the train split.
    - test_dataframe: the DataFrame that contains the test split.
    """
    in_dataframe: InArg[any]
    train_split: InArg[float]
    seed: InArg[int]

    train_dataframe: OutArg[any]
    test_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        df = self.in_dataframe.value
        seed = self.seed.value if self.seed.value else None

        train_split = self.train_split.value if self.train_split.value else 0.75

        #handle if user only provides test split

        train, test = df.randomSplit(weights=[train_split, (1 - train_split)], seed=seed)

        print("Train Split")
        train.show()
        print("Test Split")
        test.show()

        self.train_dataframe.value = train
        self.test_dataframe.value = test


@xai_component
class SparkLoadLIBSVM(Component):
    """Loads a libsvm file into a Spark dataframe.

    ##### inPorts:
    - in_sparksession: a spark session.
    - file_input: the path for the libsvm file.

    ##### outPorts:
    - out_sparksession: a spark session.
    - out_dataframe: DataFrame loaded from the libsvm.
    """    
    in_sparksession: InCompArg[any]
    file_input: InCompArg[str]
    options: InArg[any]
    out_sparksession: OutArg[any]
    out_dataframe: OutArg[any]

    def execute(self, ctx) -> None:

        spark = self.in_sparksession.value
        filepath = self.file_input.value
        options = self.options.value if self.options.value else ""

        if options:
            df = spark.read.format("libsvm").option(*options).load(filepath)
        
        else:
            df = spark.read.format("libsvm").load(filepath)

        df.show()

        self.out_sparksession.value = spark
        self.out_dataframe.value = df


@xai_component
class SparkLogisticRegression(Component):
    """Performs Spark Logistic Regression.

    ### Reference:
    - [Spark Logistic Regression](https://spark.apache.org/docs/3.3.0/ml-classification-regression.html#logistic-regression)
    
    ##### inPorts:
    - train_dataframe: a Spark dataframe.
    - family: regression mode.
        Default: `auto`. Alternative: `binomial` or `multinomial`.
    - options: Regression parameters. Supply parameters as a Dict. 
        Default: `maxIter`:10, `regParam`:0.3, `elasticNetParam`:0.8

    ##### outPorts:
    - model: trained Spark regression model.
    """    
    train_dataframe: InCompArg[any]
    family: InArg[str]
    options: InArg[dict]

    model: OutArg[any]

    def execute(self, ctx) -> None:

        from pyspark.ml.classification import LogisticRegression

        training = self.train_dataframe.value
        options = self.options.value if self.options.value else {"maxIter":10, "regParam":0.3, "elasticNetParam":0.8}
        if self.family.value:
            #You may try "multinomial" 
            options.update({"family": self.family.value})
        lr = LogisticRegression(**options)

        # Fit the model
        model = lr.fit(training)

        if (model.getFamily() == "multinomial"):
            # Print the coefficients and intercepts for logistic regression with multinomial family
            print("Multinomial coefficients: " + str(model.coefficientMatrix))
            print("Multinomial intercepts: " + str(model.interceptVector))

        else:
            # Print the coefficients and intercept for logistic regression
            print("Coefficients: " + str(model.coefficients))
            print("Intercept: " + str(model.intercept))

        self.model.value = model


@xai_component
class SparkPredict(Component):
    """Performs prediction given a Spark model and a valid corresponding dataset.
    Recommended to use with SparkSplitDataFrame component.

    ##### inPorts:
    - model: trained Spark  model.
    - test_df: Spark dataframe.

    """    
    model: InCompArg[any]
    test_df: InCompArg[any]

    def execute(self, ctx) -> None:

        model = self.model.value
        test = self.test_df.value
        test_result = model.evaluate(test)

        test_result.predictions.show()




import json

import tornado
from jupyter_server.base.handlers import APIHandler

from pathlib import Path

from xircuits.compiler import compile
import traceback


class CompileXircuitsFileRouteHandler(APIHandler):
    def __get_notebook_absolute_path__(self, path):
        return (Path(self.application.settings['server_root_dir']) / path).expanduser().resolve()

    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({"data": "This is file/compile endpoint!"}))

    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()

        input_file_path = input_data["filePath"]
        output_file_path = input_data["outPath"]

        component_python_paths = input_data["pythonPaths"]

        msg = ""

        try:
            with open(self.__get_notebook_absolute_path__(input_file_path), 'r', encoding='utf-8') as infile:
                with open(self.__get_notebook_absolute_path__(output_file_path), 'w') as outfile:
                    compile(infile, outfile, component_python_paths)

            msg = "completed"
        
        except Exception:
            msg = traceback.format_exc()
            print(msg)
            pass
        
        finally:
                
            data = {"message": msg}
            self.finish(json.dumps(data))



"""Components to perform transfer learning using pretrained models from
Tensorflow Keras, and datasets from Tensorflow Datasets.
"""
from typing import Dict

import tensorflow.keras.applications as tf_keras_applications
from tensorflow import keras
from tqdm.notebook import tqdm

from xai_components.base import Component, InArg, InCompArg, OutArg, xai_component


@xai_component(type="model")
class KerasTransferLearningModel(Component):
    """Fetch Tensorflow Keras Model by name, for transfer learning.

    ##### Reference:
    - [Keras Application
    Functions](https://www.tensorflow.org/api_docs/python/tf/keras/applications#functions_2)

    ##### inPorts:
    - base_model_name: `str`, name of model (case sensitive). The
    base_model_name must be listed under the functions
    [here](https://www.tensorflow.org/api_docs/python/tf/keras/applications#functions_2)
    - include_top: `bool`, whether to include the fully connected layers at
    the top of the network. Defaults to `True`.
    - weights: `str` pretrained weights to use. Defaults to `imagenet`.
    - input_shape: `tuple` optional shape tuple, only to be specified if
    include_top is False (otherwise the input shape has to be (224, 224, 3)
    (with channels_last data format) or (3, 224, 224) (with channels_first
    data format). It should have exactly 3 input channels, and width and
    height should be no smaller than 32. E.g. (200, 200, 3) would be one
    valid value.
    - freeze_all: `bool`, whether to freeze the weights in all layers of the
    base model. Defaults to `True`.
    - fine_tune_from: `int`, base model layer to fine tune from. Example,
    setting fine_tune_from=5 for a pretrained model with 25 layers will
    freeze only the first 5 layers. This will only take effect if freeze_all
    is set to `False`. Defaults to `0` (freeze_all=True).
    - classes: `int` number of classes to classify images into, only to be
    specified if `include_top` is `True`, and if no `weights` argument is
    specified.
    - binary: `bool` whether this model will be used for binary classification.
    Defaults o `False`.
    - classifier_activation: `str` or `callable`. The activation function to
    use on the "top" layer. Ignored unless `include_top=True`. Set
    `classifier_activation=None` to return the logits of the "top" layer.
    When loading pretrained weights, classifier_activation can only be None
    or "softmax".
    - kwargs: `dict`, optional. Passed to the model class. Please refer to the
    specific tensorflow keras model documentation for other model specific
    keyword arguments.

    ##### outPorts:
    - model: tensorflow keras model.
    """

    base_model_name: InCompArg[str]
    include_top: InArg[bool]
    weights: InArg[str]
    input_shape: InCompArg[tuple]
    freeze_all: InArg[bool]
    fine_tune_from: InArg[int]
    classes: InArg[int]
    binary: InArg[bool]
    classifier_activation: InArg[str]
    kwargs: InArg[dict]

    model: OutArg[any]

    def __init__(self):
        super().__init__()

        self.include_top.value = True
        self.weights.value = "imagenet"
        self.freeze_all.value = True
        self.fine_tune_from.value = 0
        self.classes.value = 1000
        self.binary.value = False
        self.classifier_activation.value = "softmax"
        self.kwargs.value = {}

    def execute(self, ctx):

        base_model_name = self.base_model_name.value
        # model_lookup is a dictionary containing all the available models from
        # tf.keras.applications. With key equals the model name as raw string,
        # and value equals the model function.
        # {
        #     'DenseNet121': <function DenseNet121 at 0x7f97ffa8ab80>,
        #     'DenseNet169': <function DenseNet169 at 0x7f97ffa8ac10>,
        #     'DenseNet201': <function DenseNet201 at 0x7f97ffa8aca0>,
        #     'EfficientNetB0': <function EfficientNetB0 at 0x7f97ffa9f0d0>,
        #     ...
        # }

        model_lookup = dict(
            [
                (x, getattr(tf_keras_applications, x))
                for x in dir(tf_keras_applications)
                if callable(getattr(tf_keras_applications, x))
            ]
        )
        try:
            model = model_lookup[base_model_name](
                include_top=self.include_top.value,
                weights=self.weights.value,
                input_shape=self.input_shape.value,
                classes=self.classes.value,
                **self.kwargs.value,
            )
        except KeyError as e:
            print(e)
            print("Ensure that the base model name is listed below.\n")
            print(*model_lookup.keys(), sep=", ")

        # fetch the model proprocess input layer
        model_module = getattr(tf_keras_applications, base_model_name).__module__
        model_module_name = model_module.split(".")[-1]
        model_module = getattr(tf_keras_applications, model_module_name)
        preprocess_input = getattr(model_module, "preprocess_input")

        if self.freeze_all.value:
            model.trainable = False

        if not self.freeze_all.value and self.fine_tune_from.value > 0:
            assert self.fine_tune_from.value < len(model.layers), (
                f"Please ensure that 'fine_tune_from' is lower than the "
                f"number of layers in {self.base_model_name.value} model. "
                f"{self.base_model_name.value} has {len(model.layers)} "
                f"layers, got {self.fine_tune_from.value} as the layer "
                "to 'fine_tune_from'"
            )

            model.trainable = True
            for layer in model.layers[: self.fine_tune_from.value]:
                layer.trainable = False

        if not self.include_top.value:
            assert self.input_shape.value, (
                "Please provide a valid `input_shape` if `include_top` is set "
                "to `False`. Expected a tuple of `input_shape`, e.g "
                f"`(224, 224, 3)`, got `{self.input_shape.value}`."
            )

            inputs = keras.Input(shape=self.input_shape.value)
            x = preprocess_input(inputs)
            x = model(x)
            # ------------- modify below to create your pretrained model head ------------ #
            x = keras.layers.GlobalAveragePooling2D()(x)
            # ----------------------- end of pretrained model head ----------------------- #
            if self.binary.value is True:
                outputs = keras.layers.Dense(1)(x)
            else:
                outputs = keras.layers.Dense(
                    self.classes.value, activation=self.classifier_activation.value
                )(x)

            model = keras.Model(inputs, outputs)
            print(model.summary())

        self.model.value = model


@xai_component()
class TFDataset(Component):
    """Fetch Tensorflow Dataset by name

    ##### Reference:
    - [Tensorflow Datasets
    Catalog](https://www.tensorflow.org/datasets/catalog/overview)

    ##### inPorts:
    - dataset_name: `str`, name of dataset, as listed on [Tensorflow Datasets
    catalog](https://www.tensorflow.org/datasets/catalog/overview)
    - batch_size: `int`, if set, add a batch dimension to the dataset.
    Defaults to `32`.
    - shuffle_files: `bool`, whether to shuffle the input files. Defaults to
    `False`.
    - as_supervised: `bool`, if `True`, the returned `tf.data.Dataset` will
    have a 2-tuple structure `(input, label)` according to
    `builder.info.supervised_keys`. If `False`, the default, the returned
    `tf.data.Dataset` will have a dictionary with all the features.
    - kwargs: `dict`, optional. Passed to `tfds.load`. Please refer to the
    specific tensorflow dataset documentation for other dataset specific
    keyword arguments.

    ##### outPorts:
    - all_data: `dict<key: tfds.Split, value: tf.data.Dataset>`, all available
    dataset.
    - train_data: `tf.data.Dataset`, train split if available
    - test_data: `tf.data.Dataset`, test split if available
    """

    dataset_name: InCompArg[str]
    batch_size: InArg[int]
    shuffle_files: InArg[bool]
    as_supervised: InArg[bool]
    kwargs: InArg[dict]

    all_data: OutArg[any]
    train_data: OutArg[any]
    test_data: OutArg[any]

    def __init__(self):
        super().__init__()
        self.batch_size.value = 32
        self.shuffle_files.value = False
        self.as_supervised.value = True
        self.kwargs.value = {}

    def execute(self, ctx):
        import tensorflow_datasets as tfds

        assert (
            self.dataset_name.value in tfds.list_builders()
        ), f"Please ensure that dataset_name is listed below:\n{tfds.list_builders()}"

        ds = tfds.load(
            self.dataset_name.value,
            batch_size=self.batch_size.value,
            shuffle_files=self.shuffle_files.value,
            as_supervised=self.as_supervised.value,
            **self.kwargs.value,
        )

        self.all_data.value = ds
        self.train_data.value = ds.get("train")
        self.test_data.value = ds.get("test")


@xai_component(type="train")
class TrainKerasModel(Component):
    """Trains a keras model.

    ##### inPorts:
    - model: compiled model.
    - training data: tensorflow keras model compatible dataset
    - batch_size: `int` or `None`. Number of samples per gradient update.
    - epochs: `int` number of epochs to train the model.
    - kwargs: `dict` optional. Other `tf.model.fit` arguments.

    ##### outPorts:
    - trained_model: trained tensoflow keras model.
    - training_metrics: `dict`, training metrics from training history.
    """

    model: InCompArg[any]
    training_data: InCompArg[any]
    batch_size: InCompArg[int]
    epochs: InCompArg[int]
    kwargs: InArg[dict]

    trained_model: OutArg[any]
    training_metrics: OutArg[dict]

    def __init__(self):
        super().__init__()
        self.kwargs.value = {}

    def execute(self, ctx):

        model = self.model.value
        train = model.fit(
            self.training_data.value,
            batch_size=self.batch_size.value,
            epochs=self.epochs.value,
            **self.kwargs.value,
        )

        # Set training metrics
        training_metrics = {}
        for key in train.history.keys():
            training_metrics[key] = {}
            [
                training_metrics[key].update({i + 1: v})
                for i, v in enumerate(train.history[key])
            ]

        self.trained_model.value = model
        self.training_metrics.value = training_metrics


@xai_component(type="eval")
class TFDSEvaluateAccuracy(Component):
    """Evaluate the accuracy of a Tensorflow Keras model using a Tensorflow
    dataset (`tensorflow.data.Dataset`)

    ##### inPorts:
    - model: trained tensorflow keras model.
    - eval_dataset: dataset to evaluate. Instance of `tensorflow.data.Dataset`.

    ##### outPorts:
    - metrics: `dict` model loss and accuracy.
    """

    model: InCompArg[any]
    eval_dataset: InCompArg[any]

    metrics: OutArg[Dict[str, str]]

    def execute(self, ctx):

        model_metrics_names = self.model.value.metrics_names
        model_metrics = self.model.value.evaluate(self.eval_dataset.value, verbose=0)
        metrics = dict(zip(model_metrics_names, model_metrics))
        print(metrics)

        self.metrics.value = metrics


@xai_component
class KerasModelCompiler(Component):
    """Compiles a Tensorflow Keras model.

    ##### References:
    - [Tensorflow Keras
    Optimizers](https://www.tensorflow.org/api_docs/python/tf/keras/optimizers)
    - [Tensorflow Keras
    losses](https://www.tensorflow.org/api_docs/python/tf/keras/losses)
    - [Tensorflow Keras
    Metrics](https://www.tensorflow.org/api_docs/python/tf/keras/metrics)
    - [Tensorflow Keras loss
    identifier](https://www.tensorflow.org/api_docs/python/tf/keras/losses/get#expandable-1)

    ##### inPorts:
    - model: tensorflow keras model to compile.
    - optimizer_identifier: `any`, valid tensorflow keras optimizer identifier,
    e.g, `adam` or `Adam` for default arguments, `{"class_name": "Adam",
    "config": {"learning_rate": 0.001}}` to specify keyword arguments.
    - loss_identifier: `any`, valid tensorflow keras loss identifier, e.g,
    `categorical_crossentropy` or `CategoricalCrossentropy` for a loss as a
    [function](https://www.tensorflow.org/api_docs/python/tf/keras/losses#functions_2),
    or a
    [class](https://www.tensorflow.org/api_docs/python/tf/keras/losses#classes_2),
    `{"class_name": "CategoricalCrossentropy", "config": {"from_logits": True}`
    to pass in keyword arguments. Check out the [identifier
    documentation](https://www.tensorflow.org/api_docs/python/tf/keras/losses/get#expandable-1)
    for more details.
    - metrics: `list` list of metrics to be evaluated by the model during
    training and testing. Each metric should be a string of a metric identifier,
    e.g, ['accuracy', 'mse', ... ].

    ##### outPorts:
    - compiled_model: compiled tensorflow keras model
    - model_config: `dict` model configuration.

    """

    model: InCompArg[any]
    optimizer_identifier: InCompArg[any]
    loss_identifier: InCompArg[any]
    metrics: InCompArg[list]

    compiled_model: OutArg[any]
    model_config: OutArg[dict]

    def execute(self, ctx):

        assert isinstance(
            self.model.value, keras.Model
        ), "Please pass in a tensorflow keras model"

        optimizer = keras.optimizers.get(self.optimizer_identifier.value)
        loss = keras.losses.get(self.loss_identifier.value)

        self.model.value.compile(
            optimizer=optimizer, loss=loss, metrics=self.metrics.value
        )
        self.compiled_model.value = self.model.value
        model_config = {
            "lr": self.compiled_model.value.optimizer.lr.numpy().item(),
            "optimizer_name": self.compiled_model.value.optimizer.name,
            "loss": self.compiled_model.value.loss,
        }

        self.model_config.value = model_config


@xai_component
class SaveKerasModel(Component):
    """Saves a Tensorflow Keras model.

    ##### inPorts:
    - model: tensorflow keras model to save
    - model_name: `str` name to save the model as
        
    """
    model: InCompArg[any]
    model_name: InCompArg[any]

    def execute(self, ctx):

        from pathlib import Path

        assert isinstance(
            self.model.value, keras.Model
        ), "Please pass in a tensorflow keras model"

        print(f"Saving Tensorflow Keras model to: {Path.cwd()}")
        self.model.value.save(f"{self.model_name.value}.h5")



from dataclasses import dataclass


@dataclass
class Port:
    name: str
    type: str
    dataType: str
    varName: str
    target: 'Node'
    source: 'Node'
    targetLabel: str
    sourceLabel: str
    direction: str

DYNAMIC_PORTS = ['dynalist', 'dynadict', 'dynatuple']



from xai_components.base import InArg, OutArg, InCompArg, Component, BaseComponent, xai_component, dynalist, dynatuple, dynadict

from typing import Union
@xai_component(color="red")
class HelloComponent(Component):
    """The simplest component that greets the user. 
    """
    def execute(self, ctx) -> None:
        
        #If the import is only exclusive to 1 component, it is a good practice to import inside execute()
        import os 

        creator_name = os.getlogin()
        print("Hello, " + creator_name)

@xai_component
class HelloHyperparameter(Component):
    """A component that changes the print message depending on the supplied parameter.

    ##### inPorts:
    - input_str: try connecting a Literal String or Hyperparameter String.
    """
    input_str: InArg[str]

    def execute(self, ctx) -> None:
        input_str = self.input_str.value
        print("Hello " + str(input_str))
        
@xai_component
class CompulsoryHyperparameter(Component):
    """A component that uses Compulsory inPorts. 
     Users must fill all compulsory ports to compile and run the canvas.

    ##### inPorts:
    - input_str: an optional String port.
    - comp_str: a compulsory String port.
    - comp_int: a compulsory Integer port.
    """
    input_str: InArg[str]

    #if your component requires a certain parameter to be supplied, use In-Comp(ulsory)-Argument ports.
    comp_str: InCompArg[str]
    comp_int: InCompArg[int]

    def execute(self, ctx) -> None:
        input_str = self.input_str.value
        comp_str = self.comp_str.value
        comp_int = self.comp_int.value
        print("Hello, " + str(input_str))
        print("I'm " + str(comp_str))
        print("Me " + str(comp_int))

@xai_component
class HelloListTupleDict(Component):
    """A component that accepts list, tuple, and dict data types then prints them.
    Useful for testing port type checks.

    ##### inPorts:
    - input_list: a list port.
    - input_tuple: a tuple port.
    - input_dict: a dict port.
    """
    input_list: InArg[list]
    input_tuple: InArg[tuple]
    input_dict: InArg[dict]

    def execute(self, ctx) -> None:

        #if you would like ports to have default values if user does not provide, try this way.
        input_list = self.input_list.value if self.input_list.value else ""
        input_tuple = self.input_tuple.value if self.input_tuple.value else ""
        input_dict = self.input_dict.value if self.input_dict.value else ""
        
        print( "\nDisplaying List: ")
        print(input_list) 
        print("\nDisplaying Tuple: ")
        print(input_tuple)
        print("\nDisplaying Dict: ")
        print(input_dict)

@xai_component
class MultiType(Component):
    """Component with in-port that accept multiple data types"""
    msg: InArg[Union[int, float, str]]
    
    def execute(self, ctx) -> None:
        print(str(self.msg.value))

@xai_component
class HelloContext(Component):
    """A component that showcases the usage of Xircuits Context (ctx).
    Chain multiple instances of this component to add information into the `ctx`.

    ### Reference:
    - [Xircuits Content](https://xircuits.io/docs/technical-concepts/xircuits-context)

    ##### inPorts:
    - context_dict: a dict to add to the `ctx`.
        Default: `{"new ctx": "Hello Xircuits!"}`
    """
    context_dict: InArg[dict]

    def execute(self, ctx) -> None:
        
        print(f"Current Context:\n{ctx}")
        
        context_dict = self.context_dict.value if self.context_dict.value else {"new ctx": "Hello Xircuits!"}
        ctx.update(context_dict)

        print(f"After Adding Context:\n{ctx}")
        
@xai_component
class MultiBranchComponent(BaseComponent):
    if_A: BaseComponent
    if_B: BaseComponent
    if_C: BaseComponent

    abc: InArg[str]

    def do(self, ctx) -> BaseComponent:
        if self.abc.value == "a":
            next = self.if_A
        elif self.abc.value == "b":
            next = self.if_B
        elif self.abc.value == "c":
            next = self.if_C
        else:
            next = None
        
        while next:
            next = next.do(ctx)
        try:
            return self.next
        except:
            return None
        


@xai_component
class DynaPorts(Component):
    """A component showcasing dynamic ports: `dynalist`, `dynatuple`, and `dynadict`.
    
    ##### inPorts:
    - dlist: A `dynalist` port. 
    - dtuple: A `dynatuple` port.
    - ddict: A `dynadict` port. Accepts only Literal Dicts.
    
    """
    
    dlist: InArg[dynalist]
    dtuple: InArg[dynatuple]
    ddict: InArg[dynadict]
    
    def execute(self, ctx) -> None:
        print("Printing dynalist value:")
        print(self.dlist.value)
        print("Printing dynatuple value:")
        print(self.dtuple.value)
        print("Printing dynadict value:")
        print(self.ddict.value)


import json
import re

from xircuits.compiler.node import Node
from xircuits.compiler.port import Port


class XircuitsFileParser:
    def __init__(self):
        self.traversed_nodes = {}
        self.nodes = {}
        self.links = {}

    def parse(self, input_file):

        xircuits_file = json.load(input_file)

        self.nodes = [n for n in xircuits_file['layers'] if n['type'] == 'diagram-nodes'][0]['models']
        self.links = [n for n in xircuits_file['layers'] if n['type'] == 'diagram-links'][0]['models']

        start_node = [n for n in self.nodes.values() if n['extras']['type'] == 'Start'][0]

        return self.traverse_node(start_node)

    def traverse_node(self, node):
        node_id = node['id']
        if node_id in self.traversed_nodes:
            return self.traversed_nodes[node_id]
        else:
            n = Node(
                id=node_id,
                name=node['name'],
                type=node['extras']['type'],
                file=node['extras'].get('path'),
                ports=[]
            )
            self.traversed_nodes[node_id] = n
            n.ports = self.traverse_ports(node)
            return n

    def traverse_ports(self, node):
        out = []

        for port in node['ports']:
            for linkId in port['links']:
                link = self.links[linkId]

                source_node = self.nodes[link['source']]
                target_node = self.nodes[link['target']]

                sourceLabel = [p for p in source_node['ports'] if p['id'] == link['sourcePort']][0]['label']

                # filter compulsory port [★] label from port name
                sourceLabel = re.sub(r"★", "", sourceLabel)
                targetLabel = re.sub(r"★", "", port['label'])

                p = Port(
                    name=port['name'],
                    type=link['type'],
                    varName=port['varName'],
                    dataType=port['dataType'],
                    target=self.traverse_node(target_node),
                    source=self.traverse_node(source_node),
                    targetLabel=targetLabel,
                    sourceLabel=sourceLabel,
                    direction="in" if port['in'] else "out"
                )
                out.append(p)
        return out



import argparse
from pathlib import Path
import os
import pkg_resources
import shutil
from .handlers.request_folder import request_folder, clone_from_github_url
from .handlers.request_submodule import request_submodule_library
from .compiler import compile
import subprocess
import sys
import json
import urllib.parse

def init_xircuits():
    package_name = 'xircuits'
    copy_from_installed_wheel(package_name, resource='.xircuits', dest_path='.xircuits')

def build_component_library_path(component_library_query: str) -> str:
    # ensure syntax is as xai_components/xai_library_name

    if "xai" not in component_library_query:
        component_library_query = "xai_" + component_library_query

    if "xai_components" not in component_library_query:
        component_library_query = "xai_components/" + component_library_query

    return component_library_query

def is_empty(directory):
    # will return true for uninitialized submodules
    return not os.path.exists(directory) or not os.listdir(directory)

def is_valid_url(url):
    try:
        result = urllib.parse.urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False
    
def copy_from_installed_wheel(package_name, resource="", dest_path=None):
    if dest_path is None:
        dest_path = package_name

    resource_path = pkg_resources.resource_filename(package_name, resource)
    shutil.copytree(resource_path, dest_path)

def cmd_start_xircuits(args, extra_args=[]):
    # fetch xai_components
    component_library_path = Path(os.getcwd()) / "xai_components"
    if not component_library_path.exists():
        copy_from_installed_wheel('xai_components', '', 'xai_components')

    # handler for extra jupyterlab launch options
    if extra_args:
        try:
            launch_cmd = "jupyter lab" + " " + " ".join(extra_args)
            os.system(launch_cmd)
        except Exception as e:
            print("Error in launch args! Error log:\n")
            print(e)
    else:
        os.system("jupyter lab")

def cmd_download_examples(args, extra_args=[]):
    if not os.path.exists("examples") or is_empty("examples"):
        request_folder("examples", branch=args.branch)

    if not os.path.exists("datasets") or is_empty("datasets"):
        request_folder("datasets", branch=args.branch)

def cmd_fetch_library(args, extra_args=[]):
    print(f"Fetching {args.library_name}...")

    if is_valid_url(args.library_name):
        component_library_path = clone_from_github_url(args.library_name)
        return
    
    else:
        component_library_path = build_component_library_path(args.library_name)
        
    if not Path(component_library_path).is_dir() or is_empty(Path(component_library_path)):
        request_submodule_library(component_library_path)
    else:
        print(f"{args.library_name} library already exists in {component_library_path}.")


def cmd_install_library(args, extra_args=[]):
    print(f"Installing {args.library_name}...")
    
    if is_valid_url(args.library_name):
        component_library_path = clone_from_github_url(args.library_name)

    else:
        component_library_path = build_component_library_path(args.library_name)
    
    if not Path(component_library_path).is_dir() or is_empty(Path(component_library_path)):
        request_submodule_library(component_library_path)

    requirements_file = Path(component_library_path) / "requirements.txt"

    if requirements_file.exists():
        print(f"Installing requirements for {args.library_name}...")
        subprocess.run([sys.executable, "-m", "pip", "install", "-r", str(requirements_file)], check=True)
    else:
        print(f"No requirements.txt found for {args.library_name}. Skipping installation of dependencies.")

def cmd_compile(args, extra_args=[]):
    component_paths = {}
    if args.python_paths_file:
        component_paths = json.load(args.python_paths_file)
    compile(args.source_file, args.out_file, component_python_paths=component_paths)

def get_installed_packages():
    """Return a list of installed packages."""
    result = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
    packages = {package.split('==')[0] for package in result.decode().splitlines()}
    return packages

def check_requirements_installed(installed_packages, requirements_path):
    """Check if all packages in a requirements file are installed using a given list of installed packages."""
    with open(requirements_path, 'r') as file:
        required_packages = {line.strip().split('==')[0] for line in file.readlines()}
    return required_packages.issubset(installed_packages)

def cmd_list_libraries(args, extra_args=[]):
    
    component_library_path = Path(os.getcwd()) / "xai_components"
    if not component_library_path.exists():
        copy_from_installed_wheel('xai_components', '', 'xai_components')

    print("Checking installed packages... This might take a moment.")
    installed_packages = get_installed_packages()

    # Fetch libraries from .gitmodules
    gitmodules_path = Path(".gitmodules") if Path(".gitmodules").exists() else Path(".xircuits/.gitmodules")
    submodule_paths = []
    
    if gitmodules_path.exists():
        with gitmodules_path.open() as f:
            lines = f.readlines()
            for line in lines:
                if "path = " in line:
                    submodule_paths.append(line.split("=")[-1].strip().split('/')[-1])
                    
    # Fetch libraries from xai_components/
    component_library_path = Path("xai_components")
    directories = [d.name for d in component_library_path.iterdir() if d.is_dir() and d.name.startswith("xai_")]
    non_empty_directories = [dir_name for dir_name in directories if not is_empty(component_library_path / dir_name)]

    installed_packages_dirs = []

    for dir_name in directories:
        requirements_path = component_library_path / dir_name / "requirements.txt"
        if requirements_path.exists() and check_requirements_installed(installed_packages, requirements_path):
            installed_packages_dirs.append(dir_name)

    # Crosscheck and categorize based on installed packages
    fully_installed = set(installed_packages_dirs)
    available = set(non_empty_directories) - fully_installed
    remote = [lib for lib in submodule_paths if lib not in fully_installed and lib not in available and is_empty(component_library_path / lib)]
    
    # Display
    print(f"\nFully installed component libraries({len(fully_installed)}):")
    for lib in sorted(fully_installed):
        print(f" - {lib}")

    print(f"\nAvailable component libraries({len(available)}):")
    for lib in sorted(available):
        print(f" - {lib}")

    print(f"\nRemote component libraries({len(remote)}):")
    for lib in sorted(remote):
        print(f" - {lib}")


def main():
    parser = argparse.ArgumentParser(description='Xircuits Command Line Interface', add_help=False)
    subparsers = parser.add_subparsers(dest="command")

# Adding parser for 'start' command
    start_parser = subparsers.add_parser('start', help='Start Xircuits.')
    # Add an arbitrary list of arguments. The nargs="*" means 0 or more arguments.
    # This will collect all additional arguments into a list.
    start_parser.add_argument('extra_args', nargs='*', help='Additional arguments for Xircuits launch command')
    start_parser.set_defaults(func=cmd_start_xircuits)

    # Adding parser for 'install' command
    install_parser = subparsers.add_parser('install', help='Fetch and installs a library for Xircuits.')
    install_parser.add_argument('library_name', type=str, help='Name of the library to install')
    install_parser.set_defaults(func=cmd_install_library)

    # Adding parser for 'fetch' command
    fetch_parser = subparsers.add_parser('fetch-only', help='Fetch a library for Xircuits. Does not install.')
    fetch_parser.add_argument('library_name', type=str, help='Name of the library to fetch')
    fetch_parser.set_defaults(func=cmd_fetch_library)

    # Adding parser for 'examples' command
    examples_parser = subparsers.add_parser('examples', help='Download examples for Xircuits.')
    examples_parser.add_argument('--branch', nargs='?', default="master", help='pull files from a xircuits branch')
    examples_parser.set_defaults(func=cmd_download_examples)

    # Adding parser for 'compile' command
    compile_parser = subparsers.add_parser('compile', help='Compile a Xircuits workflow file.')
    compile_parser.add_argument('source_file', type=argparse.FileType('r', encoding='utf-8'))
    compile_parser.add_argument('out_file', type=argparse.FileType('w', encoding='utf-8'))
    compile_parser.add_argument("python_paths_file", nargs='?', default=None, type=argparse.FileType('r'),
                                help="JSON file with a mapping of component name to required python path. "
                                     "e.g. {'MyComponent': '/some/path'}")
    compile_parser.set_defaults(func=cmd_compile)

    # Adding parser for 'list' command
    list_parser = subparsers.add_parser('list', help='List available component libraries for Xircuits.')
    list_parser.set_defaults(func=cmd_list_libraries)


    args, unknown_args = parser.parse_known_args()

    if hasattr(args, 'func'):
        args.func(args, unknown_args)
    else:
        valid_help_args = {"-h", "--h" "-help", "--help"}
        if any(arg in unknown_args for arg in valid_help_args):
            parser.print_help()
        else:
            # Default behavior: if no sub-command is provided, start xircuits.
            cmd_start_xircuits(args, unknown_args)

    return 0

if __name__ == '__main__':
    main()

print(
'''
======================================
__   __  ___                _ _       
\ \  \ \/ (_)_ __ ___ _   _(_) |_ ___ 
 \ \  \  /| | '__/ __| | | | | __/ __|
 / /  /  \| | | | (__| |_| | | |_\__ \\
/_/  /_/\_\_|_|  \___|\__,_|_|\__|___/
                                      
======================================
'''
)

config_path = Path(os.getcwd()) / ".xircuits"
if not config_path.exists():
    init_xircuits()



import json
import tornado
import subprocess
from jupyter_server.base.handlers import APIHandler
import platform
import os
class SparkSubmitRouteHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({"data": "This is spark/submit endpoint!"}))

    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()
        export_str = ""
        command_str = ""
        addArguments = input_data["addArgs"]
        currentPath = input_data["currentPath"]

        if platform.system() == "Windows": 
            spark_submit_cmd = "spark-submit.cmd "
        else: 
            spark_submit_cmd = "spark-submit "

        for key, value in addArguments.items():
            if key == 'export':
                export_str = value
            elif key == 'command':
                command_str = value
        spark_submit_str= command_str + " " + currentPath
        # env = {
        #     **os.environ
        # }
        # if export_str != 'null':
        #     process=subprocess.Popen(spark_submit_str,stdout=subprocess.PIPE,stderr=subprocess.PIPE, env=export_str,universal_newlines=True, shell=True)
        #     stdout = process.communicate()
        # else:
        process=subprocess.Popen(spark_submit_str,stdout=subprocess.PIPE,stderr=subprocess.PIPE,universal_newlines=True, shell=True)
        stdout,stderr = process.communicate()

        data = {
            "stderr": stderr,
            "stdout": stdout
        }

        self.finish(json.dumps(data))


import json

import requests
import tornado
from jupyter_server.base.handlers import APIHandler

from .config import get_config


class DebuggerRouteHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({"data": "This is /debug/enable endpoint!"}))

    @tornado.web.authenticated
    def post(self):
        input_data = self.get_json_body()

        cfg = get_config()
        port = str(cfg['SERVER']['IP_ADD'] + ":" + cfg['SERVER']['PORT'])

        output_content = ""
        if input_data["command"] == "run":
            output = requests.get(port + "/run") # next node

        elif input_data["command"] == "continue":
            output = requests.get(port +"/continue")

        elif input_data["command"] == "get/output":
            output = requests.get(port + "/output")

        elif input_data["command"] == "clear":
            output = requests.get(port + "/clear")

        elif input_data["command"] == "get_run":
            output = requests.get(port + "/execute")

        elif input_data["command"] == "clear_run":
            output = requests.get(port + "/clear_execution")

        elif input_data["command"] == "terminate":
            output = requests.get(port + "/terminate")

        else:
            output = ""

        try:
            if output != "":
                encoding = 'utf-8'
                output_content = output.content.decode(encoding)
        except:
            output_content = "Incorrect endpoint"

        data = {"output": output_content}

        self.finish(json.dumps(data))


from xai_components.base import InArg, InCompArg, OutArg, Component, xai_component

import os
import sys

import torch
from torch import nn

@xai_component
class LoadTorchVisionDataset(Component):
    """Loads a Torch Vision dataset.
    
    ### Reference:
    - [Pytorch Vision Datasets](https://pytorch.org/vision/stable/datasets.html#built-in-datasets)

    ##### inPorts:
    - dataset_name: Loads a valid Pytorch Vision dataset, downloads from the 
    repository if not present in `dataset_dir`. 
    - dataset_dir: Path to save downloaded Pytorch Vision dataset.
        Default: `./data`.

    ##### outPorts:
    - training_data: train split of the dataset.
    - test_data: test split of the dataset.
    """     

    dataset_name: InCompArg[str]
    dataset_dir: InArg[str]

    training_data: OutArg[torch.utils.data.Dataset]
    test_data: OutArg[torch.utils.data.Dataset]


    def execute(self,ctx) -> None:

        from torchvision import datasets
        from torchvision.transforms import ToTensor

        dataset_dir = self.dataset_dir.value if self.dataset_dir.value else "data"

        print("Downloading " + self.dataset_name.value + " to " + os.path.abspath(dataset_dir))
        # Download training data from open datasets.
        training_data = getattr(datasets, self.dataset_name.value)(
            root=dataset_dir,
            train=True,
            download=True,
            transform=ToTensor(),
        )

        # Download test data from open datasets.
        test_data = getattr(datasets, self.dataset_name.value)(
            root=dataset_dir,
            train=False,
            download=True,
            transform=ToTensor(),
        )

        self.training_data.value = training_data
        self.test_data.value = test_data

@xai_component
class TorchDataLoader(Component):
    """Creates data iterators from torch datasets. 
    To be used with `LoadTorchVisionDataset` component.

    ##### inPorts:
    - training_data: Torch Dataset of training data.
    - test_data: Torch Dataset of test data.
    - batch_size: batch size to iterate though.
        Default: 64

    ##### outPorts:
    - train_dataloader: Dataloader instance for training data.
    - test_dataloader: Dataloader instance for test data.
    """

    training_data: InCompArg[torch.utils.data.Dataset]
    test_data: InCompArg[torch.utils.data.Dataset]
    batch_size: InArg[int]

    train_dataloader: OutArg[torch.utils.data.DataLoader]
    test_dataloader: OutArg[torch.utils.data.DataLoader]

    def execute(self,ctx) -> None:

        from torch.utils.data import DataLoader

        batch_size = self.batch_size.value if self.batch_size.value else 64

        # Create data loaders.
        train_dataloader = DataLoader(self.training_data.value, batch_size=batch_size)
        test_dataloader = DataLoader(self.test_data.value, batch_size=batch_size)

        for X, y in test_dataloader:
            print(f"Shape of X [N, C, H, W]: {X.shape}")
            print(f"Shape of y: {y.shape} {y.dtype}")
            break

        self.train_dataloader.value = train_dataloader
        self.test_dataloader.value = test_dataloader

        

@xai_component
class ExampleTorchModelConfig(Component):
    """Creates example Torch Model config.
    ##### outPorts:
    - model: torch.nn instance that expects a 28*28 input.
    - loss_fn: nn.CrossEntropyLoss()
    - optimizer: torch.optim.SGD(model.parameters(), lr=1e-3)
    """

    model_config: OutArg[nn.Module]
    loss_fn: OutArg[any]
    optimizer: OutArg[any]

    def execute(self,ctx) -> None:
        
        # Define model
        class NeuralNetwork(nn.Module):
            def __init__(self):
                super(NeuralNetwork, self).__init__()
                self.flatten = nn.Flatten()
                self.linear_relu_stack = nn.Sequential(
                    nn.Linear(28*28, 512),
                    nn.ReLU(),
                    nn.Linear(512, 512),
                    nn.ReLU(),
                    nn.Linear(512, 10)
                )

            def forward(self, x):
                x = self.flatten(x)
                logits = self.linear_relu_stack(x)
                return logits

        # Get cpu or gpu device for training.
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")
        model = NeuralNetwork().to(device)
        print(model)

        loss_fn = nn.CrossEntropyLoss()
        optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)

        self.model_config.value = model
        self.loss_fn.value = loss_fn
        self.optimizer.value = optimizer

@xai_component
class TrainTorchModel(Component):
    """Trains a Torch model instance from a dataloader. 

    ##### inPorts:
    - train_dataloader: torch dataloader util instance. Ideally from `TorchDataLoader`.
    - model: torch.nn instance.
    - loss_fn: torch.nn loss function.
    - optimizer: torch model optimizer.
    - epochs: training epochs. 
        Default: `5`.

    ##### outPorts:
    - trained_model: trained torch.nn instance. 
    """
    train_dataloader: InCompArg[torch.utils.data.DataLoader]
    model_config: InCompArg[nn.Module]
    loss_fn: InCompArg[any]
    optimizer: InCompArg[any]
    epochs: InArg[int]

    trained_model: OutArg[nn.Module]

    def execute(self,ctx) -> None:

        dataloader = self.train_dataloader.value
        model = self.model_config.value
        loss_fn = self.loss_fn.value
        optimizer = self.optimizer.value
        epochs = self.epochs.value if self.epochs.value else 5

        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")

        for t in range(epochs):
            print(f"\nEpoch {t+1}\n-------------------------------")
            
            size = len(dataloader.dataset)
            model.train()
            for batch, (X, y) in enumerate(dataloader):
                X, y = X.to(device), y.to(device)

                # Compute prediction error
                pred = model(X)
                loss = loss_fn(pred, y)

                # Backpropagation
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                if batch % 100 == 0:
                    loss, current = loss.item(), batch * len(X)
                    print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")

        self.trained_model.value = model
@xai_component
class TestTorchModel(Component):
    """Tests a Torch model instance from a dataloader. 

    ##### inPorts:
    - test_dataloader: torch dataloader util instance. Ideally from `TorchDataLoader`.
    - model: torch.nn instance. Ideally trained.
    - loss_fn: torch.nn loss function.
    """
    model: InCompArg[nn.Module]
    test_dataloader: InCompArg[torch.utils.data.DataLoader]

    loss_fn: InCompArg[any]
    

    def execute(self,ctx) -> None:
        
        dataloader = self.test_dataloader.value
        model = self.model.value
        loss_fn = self.loss_fn.value

        # Get cpu or gpu device for training.
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")

        size = len(dataloader.dataset)
        num_batches = len(dataloader)
        model.eval()
        test_loss, correct = 0, 0
        with torch.no_grad():
            for X, y in dataloader:
                X, y = X.to(device), y.to(device)
                pred = model(X)
                test_loss += loss_fn(pred, y).item()
                correct += (pred.argmax(1) == y).type(torch.float).sum().item()
        test_loss /= num_batches
        correct /= size
        print(f"Test Error: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")
        
@xai_component
class SaveTorchModelState(Component):
    """Saves a Torch model's trained state.

    ##### inPorts:
    - model: trained torch.nn instance. 
    - model_path: the path/filename to be saved as. `.pth` will be appended.
        Default: .xircuits filename + .pth.
    """
    model: InCompArg[nn.Module]
    model_path: InArg[str]

    def execute(self,ctx) -> None:
        
        model_path = self.model_path.value if self.model_path.value else os.path.splitext(sys.argv[0])[0] + ".pth"
        torch.save(self.model.value.state_dict(), model_path)

        print("Saved PyTorch Model State to " + model_path)

@xai_component
class LoadTorchModelState(Component):
    """Loads a Torch model's state from a previously saved .pth.

    ##### inPorts:
    - model_config: torch.nn config instance. 
    - model_path: the saved model path.

    ##### outPorts:
    - loaded_model: torch.nn model with loaded state. 
    """
    model_config: InCompArg[nn.Module]
    model_path: InCompArg[str]

    loaded_model: OutArg[nn.Module]

    def execute(self,ctx) -> None:

        # Get cpu or gpu device for training.
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")

        model = self.model_config.value.to(device)
        model.load_state_dict(torch.load(self.model_path.value))

        self.loaded_model.value = model

@xai_component
class TorchModelPredict(Component):
    """Performs a prediction given a Torch model, test_data split, and a class list.

    ##### inPorts:
    - model: trained torch.nn instance.
    - test_data: a torch dataset split. Ideally from `LoadTorchVisionDataset`.
    - class_list: list of classes model was trained on.
    """
    model: InCompArg[nn.Module]
    test_data: InCompArg[any]
    class_list: InCompArg[list]

    def execute(self,ctx) -> None:

        test_data = self.test_data.value
        classes = self.class_list.value

        x, y = test_data[0][0], test_data[0][1]

        model = self.model.value
        model.eval()

        with torch.no_grad():
            pred = model(x)
            predicted, actual = classes[pred[0].argmax(0)], classes[y]
            print(f'Predicted: "{predicted}", Actual: "{actual}"')



@xai_component
class Image2TorchTensor(Component):
    """Converts an image loaded from path to a torch tensor.

    ##### inPorts:
    - img_path: image path.
    - resize: tuple with desired tensor dimension.
        If not provided, will use original image dimensions.

    ##### outPorts:
    - tensor: a torch tensor instance.
    """
    img_path: InCompArg[str]
    resize: InArg[tuple]

    tensor: OutArg[torch.Tensor]

    def execute(self,ctx) -> None:

        from torchvision import transforms
        from PIL import Image

        # mnistFashion expects (1, 28, 28)
        img = Image.open(self.img_path.value).convert('L')
        
        print("Size of the Original image: ", img.size)

        if self.resize.value:
            transform = transforms.Resize(size = (self.resize.value))
            img = transform(img)
            print("Size of the image after resize: ", img.size)

        convert_tensor = transforms.ToTensor()
        tensor = convert_tensor(img)

        print("Size of the tensor: ", tensor.size())

        self.tensor.value = tensor

@xai_component
class TorchModelPredictFromTensor(Component):
    """Performs a prediction given a Torch model, tensor, and a class list.
    Ideally to be used with `Image2TorchTensor`.

    ##### inPorts:
    - model: trained torch.nn instance.
    - test_data: a torch tensor.
    - class_list: list of classes model was trained on.
    """
    model: InCompArg[nn.Module]
    tensor: InCompArg[torch.Tensor]
    class_list: InCompArg[list]

    def execute(self,ctx) -> None:

        # Get cpu or gpu device for training.
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")

        classes = self.class_list.value        
        x = self.tensor.value.to(device)
        model = self.model.value

        model.eval()

        with torch.no_grad():
            pred = model(x)
            predicted = classes[pred[0].argmax(0)]
            print(f'Predicted: "{predicted}"')



import json
import os
import pathlib
import sys
import ast
from itertools import chain
import traceback

import tornado
from jupyter_server.base.handlers import APIHandler
import platform

from .config import get_config

DEFAULT_COMPONENTS_PATHS = [
    os.path.join(os.path.dirname(__file__), "..", "..", "xai_components"),
    "xai_components",
    os.path.expanduser("~/xai_components"),
    os.environ.get("XPIPES_COMPONENTS_DIR")
]

# Get the default components from here for now
# A better place may be a config file, or turning them into real components
# A good point in time to do that, would be when the python compilation step
# gets refactored
DEFAULT_COMPONENTS = {
    1: { "name": "Get Argument String Name", "returnType": "string","color":"lightpink"},
    2: { "name": "Get Argument Integer Name", "returnType": "int","color":"blue"},
    3: { "name": "Get Argument Float Name", "returnType": "float","color":"green"},
    4: { "name": "Get Argument Boolean Name", "returnType": "boolean","color":"red"},
    5: { "name": "Literal String", "returnType": "string","color":"lightpink"},
    6:{ "name": "Literal Integer", "returnType": "int","color":"blue"},
    7:{ "name": "Literal Float", "returnType": "float","color":"green"},
    8:{ "name": "Literal True", "returnType": "boolean","color":"red"},
    9:{ "name": "Literal False", "returnType": "boolean","color":"red"},
    10:{ "name": "Literal List", "returnType": "list","color":"yellow"},
    11:{ "name": "Literal Tuple", "returnType": "tuple","color":"purple"},
    12:{ "name": "Literal Dict", "returnType": "dict","color":"orange"},
    13:{ "name": "Literal Secret", "returnType": "secret","color":"black"},
    14:{ "name": "Literal Chat", "returnType": "chat","color":"green"},

    # Comment this first since we don't use it
    # 1: { "name": "Math Operation", "returnType": "math"},
    # 2: { "name": "Convert to Aurora", "returnType": "convert"},
    # 7: { "name": "Debug Image", "returnType": "debug"},
    # 8: { "name": "Reached Target Accuracy", "returnType": "enough"},
}

COLOR_PALETTE = [
    "rgb(192,255,0)",
    "rgb(0,102,204)",
    "rgb(255,153,102)",
    "rgb(255,102,102)",
    "rgb(15,255,255)",
    "rgb(255,204,204)",
    "rgb(153,204,51)",
    "rgb(255,153,0)",
    "rgb(255,204,0)",
    "rgb(204,204,204)",
    "rgb(153,204,204)",
    "rgb(153,0,102)",
    "rgb(102,51,102)",
    "rgb(153,51,204)",
    "rgb(102,102,102)",
    "rgb(255,102,0)",
    "rgb(51,51,51)"
]

GROUP_GENERAL = "GENERAL"
GROUP_ADVANCED = "ADVANCED"

def remove_prefix(input_str, prefix):
    prefix_len = len(prefix)
    if input_str[0:prefix_len] == prefix:
        return input_str[prefix_len:]
    else:
        return input_str

def read_orig_code(node: ast.AST, lines):
    line_from = node.lineno - 1
    col_from = node.col_offset

    line_to = node.end_lineno - 1
    col_to = node.end_col_offset

    if line_from == line_to:
        line = lines[line_from]
        return line[col_from:col_to]
    else:
        start_line = lines[line_from][col_from:]
        between_lines = lines[(line_from+1):line_to]
        end_line = lines[line_to][col_to]
        return "\n".join(chain([start_line], between_lines, [end_line]))


class ComponentsRouteHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        components = []
        error_msg = ""

        for id, c in DEFAULT_COMPONENTS.items():
            components.append({
                "task": c["name"],
                "header": GROUP_GENERAL,
                "category": GROUP_GENERAL,
                "variables": [],
                "type": c["returnType"],
                "color":c.get('color') or None    
            })

        default_paths = set(pathlib.Path(p).expanduser().resolve() for p in sys.path)

        visited_directories = []
        for directory_string in self.get_component_directories():
            if directory_string is not None:
                directory = pathlib.Path(directory_string).absolute()
                if directory.exists() \
                        and directory.is_dir() \
                        and not any(pathlib.Path.samefile(directory, d) for d in visited_directories):
                    visited_directories.append(directory)
                    python_files = directory.rglob("xai_*/*.py")

                    python_path = directory.expanduser().resolve()

                    if python_path.parent in default_paths:
                        python_path = None

                    try:
                        components.extend(chain.from_iterable(self.extract_components(f, directory, python_path) for f in python_files))
                    except Exception:
                        error_msg = traceback.format_exc()
                        pass
                    finally:
                        components.extend(chain.from_iterable(self.extract_components(f, directory, python_path) for f in python_files))


        components = list({(c["header"], c["task"]): c for c in components}.values())

        # Set up component colors according to palette
        for idx, c in enumerate(components):
            if c.get("color") is None:
                c["color"] = COLOR_PALETTE[idx % len(COLOR_PALETTE)]

        data = {"components": components,
                "error_msg" : error_msg}

        self.finish(json.dumps(data))
        
    def get_component_directories(self):
        paths = list(DEFAULT_COMPONENTS_PATHS)
        paths.append(get_config().get("DEV", "BASE_PATH"))
        return paths

    def extract_components(self, file_path, base_dir, python_path):
        with open(file_path) as f:
            lines = f.readlines()
        
        parse_tree = ast.parse(file_path.read_text(), file_path)
        # Look for top level class definitions that are decorated with "@xai_component"
        is_xai_component = lambda node: isinstance(node, ast.ClassDef) and \
                                        any((isinstance(decorator, ast.Call) and decorator.func.id == "xai_component") or \
                                            (isinstance(decorator, ast.Name) and decorator.id == "xai_component")
                                            for decorator in node.decorator_list)

        return [self.extract_component(node, file_path.relative_to(base_dir), lines, python_path)
                for node in parse_tree.body if is_xai_component(node)]

    def extract_component(self, node: ast.ClassDef, file_path, file_lines, python_path):
        name = node.name

        keywords = {kw.arg: kw.value.value for kw in chain.from_iterable(decorator.keywords
                        for decorator in node.decorator_list
                        if isinstance(decorator, ast.Call) and decorator.func.id == "xai_component")}

        # Group Name for Display
        category = remove_prefix(file_path.parent.name, "xai_").upper()

        is_arg = lambda n: isinstance(n, ast.AnnAssign) and \
                                           isinstance(n.annotation, ast.Subscript) and \
                                           n.annotation.value.id in ['InArg', 'InCompArg', 'OutArg']

        is_flow_arg = lambda n: isinstance(n, ast.AnnAssign) and \
                                            isinstance(n.annotation, ast.Name) and \
                                            n.annotation.id in ['BaseComponent']

        python_version = platform.python_version_tuple()

        variables = []
        for v in (node.body):
            if is_flow_arg(v):
                variables.append({
                    "name": v.target.id,
                    "kind": v.annotation.id,
                })
                continue
            elif is_arg(v):
                variables.append({
                    "name": v.target.id,
                    "kind": v.annotation.value.id,
                    "type": read_orig_code(v.annotation.slice.value if int(python_version[1]) == 8 else v.annotation.slice, file_lines)
                })
                continue

        docstring = ast.get_docstring(node)
        lineno = [
            {
                "lineno": node.lineno,
                "end_lineno": node.end_lineno
            }
        ]

        output = {
            "class": name,
            "package_name": ("xai_components." if python_path is None else "") + file_path.as_posix().replace("/", ".")[:-3],
            "python_path": str(python_path) if python_path is not None else None,
            "abs_file_path": os.path.join(str(python_path), str(file_path)) if python_path is not None else None,
            "file_path": "xai_components/" + (file_path.as_posix()[:-3] + ".py" if platform.system() == "Windows" else str(file_path)),
            "task": name,
            "header": GROUP_ADVANCED,
            "category": category,
            "type": "debug",
            "variables": variables,
            "docstring": docstring,
            "lineno" : lineno
        }
        output.update(keywords)

        return output



from xai_components.base import InArg, OutArg, InCompArg, Component, xai_component

@xai_component
class XGBoostClassifier(Component):
    """
    Trains an XGBoost classifier. Default will use binary logistic.

    #### Reference:
    - [XGBoost](https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBClassifier)

    ##### inPorts:
    - X_train: The training data.
    - y_train: The target variable for the training data.
    - n_estimators: Number of gradient boosted trees. Equivalent to the number of boosting rounds (default is 2).
    - max_depth: Maximum tree depth for base learners (default is 2).
    - learning_rate: Boosting learning rate (default is 1).
    - objective: Specify the learning task and the corresponding learning objective (default is 'binary:logistic').

    ##### outPorts:
    - bst: The trained XGBoost model.
    """

    X_train: InCompArg[any]
    y_train: InCompArg[any]
    n_estimators: InArg[int]
    max_depth: InArg[int]
    learning_rate: InArg[float]
    objective: InArg[str]
    bst: OutArg[any]

    def __init__(self):
        super().__init__()
        self.n_estimators.value = 2
        self.max_depth.value = 2
        self.learning_rate.value = 1
        self.objective.value = 'binary:logistic'

    def execute(self, ctx) -> None:
        from xgboost import XGBClassifier

        self.bst.value = XGBClassifier(n_estimators=self.n_estimators.value, max_depth=self.max_depth.value, 
                                    learning_rate=self.learning_rate.value, objective=self.objective.value)
        self.bst.value.fit(self.X_train.value, self.y_train.value)


@xai_component
class XGBoostPredict(Component):
    """
    Makes predictions using a trained XGBoost classifier and optionally evaluates the accuracy of those predictions.

    #### Reference:
    - [XGBoost](https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBClassifier)

    ##### inPorts:
    - bst: The trained XGBoost model.
    - X_test: The testing data.
    - y_test: The target variable for the testing data. If provided, the accuracy of the predictions is evaluated.

    ##### outPorts:
    - preds: The model's predictions.
    - accuracy: The accuracy of the model's predictions, if y_test was provided.
    """

    bst: InCompArg[any]
    X_test: InCompArg[any]
    y_test: InArg[any]
    preds: OutArg[any]
    accuracy: OutArg[float]

    def execute(self, ctx) -> None:
        from sklearn.metrics import accuracy_score

        self.preds.value = self.bst.value.predict(self.X_test.value)

        if self.y_test.value is not None:
            self.accuracy.value = accuracy_score(self.y_test.value, self.preds.value)
            print(f"Accuracy: {self.accuracy.value * 100:.2f}%")
        else:
            self.accuracy.value = None


from xai_components.base import InArg, OutArg, InCompArg, Component, xai_component

import os
import sys
from pathlib import Path
import time

@xai_component
class Print(Component):
    msg: InArg[any]
    
    def execute(self, ctx) -> None:
        print(str(self.msg.value))
        
@xai_component
class ConcatString(Component):
    a: InArg[str]
    b: InArg[str]
    out: OutArg[str]

    def execute(self, cts) -> None:
        self.out.value = self.a.value + self.b.value
    
@xai_component
class ZipDirectory(Component):
    """Zips a directory.
    
    ##### inPorts:
    - zip_fn: the Zip filename.
        Default: .xircuits canvas name.
    - dir_name: the directory to be zipped.
    - include_dir: bundle the specified directory in the zip if True.
    
        #### EG:

        Creating Bar.zip which zips the dir `Foo` with the structure:
        ```  
        Foo
        |-- file1
        |-- file2
        ```

        if `include_dir` is `True`:
        ```
        Bar.zip
        |-- Foo
            |-- file1
            |-- file2
        ```

        if `include_dir` is `False`:
        ```
        Bar.zip
        |-- file1
        |-- file2
        ```

    """
    zip_fn: InArg[str]
    dir_name: InCompArg[str]
    include_dir: InArg[bool]

    def execute(self, ctx) -> None:
        
        from zipfile import ZipFile
        from tqdm import tqdm

        zip_fn = self.zip_fn.value if self.zip_fn.value else Path(sys.argv[0]).stem
        dir_name = self.dir_name.value

        assert os.path.isdir(dir_name), "invalid dir_name!"

        if os.path.splitext(zip_fn)[-1] != ".zip":
            zip_fn = zip_fn + ".zip"

        if not Path(zip_fn).is_file():
            print(zip_fn + " created at " + os.getcwd()) 
            zipObj = ZipFile(zip_fn, 'w')

        else:
            print(zip_fn + " updated at " + os.getcwd()) 
            zipObj = ZipFile(zip_fn,'a')

        for root, dirs, files in tqdm(list(os.walk(dir_name))):
            
            #chop off root dir
            if self.include_dir.value == False:
                length = len(dir_name)
                dirs = root[length:]
                
            for filename in files:
                
                if self.include_dir.value == False:
                    zipObj.write(os.path.join(root, filename), os.path.join(dirs, filename))
                
                else:
                    zipObj.write(os.path.join(root, filename))

        zipObj.close()        
       
@xai_component
class DeleteFile(Component):
    """Deletes a file.
    
    ##### inPorts:
    - filename: path to file to be deleted.
    """
    filename: InCompArg[str]

    def execute(self, ctx) -> None:

        filename = self.filename.value if self.filename.value else None

        if os.path.exists(filename):
          os.remove(filename)
        else:
          print(filename + " does not exist.") 

@xai_component(color="green")
class TimerComponent(Component):
    """Chain multiple instances of this component to measure elapsed time.
    
    ##### inPorts:
    - in_timer: if provided will measure the elapsed time since last called.
        if not provided, will start a new timer.
    - timer_message: a log message to be printed. 

    ##### outPorts:
    - out_timer: passes current timer to the next TimerComponent. 
    - elapsed_time: the elapsed time in seconds.
    """
    in_timer: InArg[time.time]
    timer_message: InArg[str]
    
    out_timer: OutArg[time.time]
    elapsed_time: OutArg[int]
    
    def execute(self, ctx) -> None:
                
        current_time = time.time()
        timer_message = self.timer_message.value if self.timer_message.value else "current timer component"
        
        elapsed_time = 0
        
        if self.in_timer.value:
                elapsed_time = current_time - self.in_timer.value
                print("The elapsed time for " + timer_message + " is " + str(elapsed_time) + " seconds.")
        else:
            print("Starting new timer.")
        
        self.out_timer.value = current_time
        self.elapsed_time.value = elapsed_time

        
@xai_component(color="green")
class SleepComponent(Component):
    """Pauses the python process.
    
    ##### inPorts:
    - sleep_timer: the number of seconds to pause.
        Default `5.0` seconds.
    """    
    sleep_timer: InArg[float]

    def execute(self, ctx) -> None:
        
        sleep_timer = self.sleep_timer.value if self.sleep_timer.value else 5.0
        print("Sleeping for " + str(sleep_timer) + " seconds.")
        time.sleep(sleep_timer)


import ast
import itertools
import re
import json
import sys
from .port import DYNAMIC_PORTS

if sys.version_info >= (3, 9):
    from ast import unparse
else:
    import io
    from xircuits.compiler.vendor.unparse import Unparser


    def unparse(parsed):
        f = io.StringIO()
        Unparser(parsed, f)
        return f.getvalue()


class CodeGenerator:
    def __init__(self, graph, component_python_paths):
        self.graph = graph
        self.component_python_paths = component_python_paths

    def generate(self, filelike):
        module_body = list(itertools.chain(
            self._generate_python_path(),
            self._generate_fixed_imports(),
            self._generate_component_imports(),
            self._generate_main(),
            self._generate_trailer()
        ))

        with open(filelike.name, 'w', encoding='utf-8') as f:
            f.write(unparse(ast.Module(body=module_body, type_ignores=[])))

    def _generate_python_path(self):
        fixed_code = """
import sys        
"""
        code = ast.parse(fixed_code).body
        nodes = self._build_node_set()

        needed_paths = set()
        for node in nodes:
            if node.name in self.component_python_paths:
                needed_paths.add(self.component_python_paths[node.name])

        if len(needed_paths) == 0:
            return []

        for p in needed_paths:
            tpl = ast.parse('sys.path.append("value")')
            tpl.body[0].value.args[0].value = p
            code.extend(tpl.body)
        return code

    def _generate_fixed_imports(self):
        fixed_imports = """
from argparse import ArgumentParser
from xai_components.base import SubGraphExecutor

"""
        return ast.parse(fixed_imports).body

    def _generate_component_imports(self):
        nodes = self._build_node_set()
        unique_components = list(set((n.name, n.file) for n in nodes if n.file is not None))
        unique_components.sort(key=lambda x: x[1])
        unique_modules = itertools.groupby(unique_components, lambda x: x[1])

        imports = []
        for (file, components) in unique_modules:
            module = '.'.join(file[:-3].split('/'))

            imports.append(
                ast.parse("from %s import %s" % (module, ",".join(c[0] for c in components)))
            )

        return imports

    def _generate_main(self):
        main = ast.parse("""
def main(args):
    ctx = {}
    ctx['args'] = args
""""").body[0]

        nodes = self._build_node_set()
        component_nodes = [n for n in nodes if n.file is not None]
        named_nodes = dict((n.id, "c_%s" % idx) for idx, n in enumerate(component_nodes))

        code = []

        # Instantiate all components
        code.extend([
            ast.parse("%s = %s()" % (named_nodes[n.id], n.name)) for n in component_nodes
        ])

        # Set up component argument links
        for node in component_nodes:
            # Handle argument connections
            for port in (p for p in node.ports if
                         p.direction == 'in' and p.type == 'triangle-link' and p.source.name.startswith('Argument ')):
                # Unfortunately, we don't have the information anywhere else and updating the file format isn't an option at the moment

                pattern = re.compile(r'^Argument \(.+?\): (.+)$')
                arg_name = pattern.match(port.source.name).group(1)

                assignment_target = "%s.%s.value" % (
                    named_nodes[port.target.id],
                    port.targetLabel
                )
                assignment_source = "args.%s" % arg_name
                tpl = ast.parse("%s = %s" % (assignment_target, assignment_source))
                code.append(tpl)

            # Handle regular connections
            for port in (p for p in node.ports if p.direction == 'in' and p.type != 'triangle-link' and p.dataType not in DYNAMIC_PORTS):
                assignment_target = "%s.%s" % (
                    named_nodes[port.target.id],
                    port.targetLabel
                )

                if port.source.id not in named_nodes:

                    # Literal
                    assignment_target += ".value"
                    tpl = ast.parse("%s = 1" % (assignment_target))
                    if port.source.type == "string":
                        value = port.sourceLabel
                    elif port.source.type == "list":
                        value = json.loads("[" + port.sourceLabel + "]")
                    elif port.source.type == "dict":
                        value = json.loads("{" + port.sourceLabel + "}")
                    elif port.source.type == "secret":
                        value = port.sourceLabel
                    elif port.source.type == "chat":
                        value = json.loads(port.sourceLabel)
                    elif port.source.type == "tuple":
                        value = eval(port.sourceLabel)
                        if not isinstance(value, tuple):
                            # handler for single entry tuple
                            value = (value,)
                    else:
                        value = eval(port.sourceLabel)
                    tpl.body[0].value.value = value
                else:
                    assignment_source = "%s.%s" % (
                        named_nodes[port.source.id],
                        port.sourceLabel
                    )
                    tpl = ast.parse("%s = %s" % (assignment_target, assignment_source))
                code.append(tpl)

            # Handle dynamic connections
            dynaports = [p for p in node.ports if p.direction == 'in' and p.type != 'triangle-link' and p.dataType in DYNAMIC_PORTS]

            ports_by_varName = {}  

            # Group ports by varName
            for port in dynaports:
                if port.varName not in ports_by_varName:  
                    ports_by_varName[port.varName] = []
                ports_by_varName[port.varName].append(port)

            # Iterate through grouped ports and append values
            for varName, ports in ports_by_varName.items():
                appended_list = []
                merged_dict = {}
                convert_to_tuple = False 

                for port in ports:
                    if port.source.id not in named_nodes:
                        if port.source.type == "string":
                            value = port.sourceLabel
                        elif port.source.type == "list":
                            value = json.loads("[" + port.sourceLabel + "]")
                        elif port.source.type == "dict":
                            value = json.loads("{" + port.sourceLabel + "}")
                        elif port.source.type == "tuple":
                            value = list(eval(port.sourceLabel))
                        else:
                            value = eval(port.sourceLabel)
                        
                    else:
                        value = "%s.%s" % (named_nodes[port.source.id], port.sourceLabel)

                    if port.dataType == 'dynadict':
                        merged_dict.update(value)
                    else:
                        appended_list.append(value)
                    
                    if port.dataType == 'dynatuple':
                        convert_to_tuple = True

                # Create a single AST node for each unique varName and append to code
                assignment_target = "%s.%s.value" % (named_nodes[ports[0].target.id], ports[0].varName)
                
                assignment_value = ''

                if merged_dict:
                    assignment_value = repr(merged_dict)
                elif appended_list:
                    if convert_to_tuple:
                        assignment_value = repr(tuple(appended_list))
                    else:
                        assignment_value = repr(appended_list)
                
                tpl = ast.parse("%s = %s" % (assignment_target, assignment_value))
                code.append(tpl)

        # Set up control flow
        for node in component_nodes:
            has_next = False
            for port in (p for p in node.ports if p.direction == "out" and p.type == "triangle-link"):
                has_next = True
                if port.name == "out-0":
                    assignment_target = "%s.next" % named_nodes[port.source.id]
                    assignment_source = named_nodes[port.target.id] if port.target.id in named_nodes else None
                    code.append(
                        ast.parse("%s = %s" % (assignment_target, assignment_source))
                    )
                elif port.name.startswith("out-flow-"):
                    assignment_target = "%s.%s" % (named_nodes[port.source.id], port.name[len("out-flow-"):])
                    assignment_source = "SubGraphExecutor(%s)" % named_nodes[
                        port.target.id] if port.target.id in named_nodes else None
                    code.append(
                        ast.parse("%s = %s" % (assignment_target, assignment_source))
                    )
            if not has_next:
                assignment_target = "%s.next" % named_nodes[node.id]
                code.append(
                    ast.parse("%s = None" % assignment_target)
                )

        trailer = """
next_component = %s
while next_component:
    next_component = next_component.do(ctx)        
        """ % (named_nodes[self.graph.ports[0].target.id])
        code.append(ast.parse(trailer))

        main.body.extend(code)
        return [main]

    def _build_node_set(self):
        nodes = set()
        node_queue = [self.graph]
        while len(node_queue) != 0:
            current_node = node_queue.pop()
            nodes.add(current_node)
            for port in current_node.ports:
                if port.target not in nodes:
                    node_queue.append(port.target)
                if port.source not in nodes:
                    node_queue.append(port.source)
        return nodes

    def _generate_trailer(self):
        code = """
if __name__ == '__main__':
    main(parser.parse_args())
    print("\\nFinished Executing")        
        """
        body = ast.parse(code).body[0]
        arg_parsing = self._generate_argument_parsing()
        arg_parsing.extend(body.body)
        body.body = arg_parsing
        return [body]

    def _generate_argument_parsing(self):
        # Unfortunately, we don't have the information anywhere else and updating the file format isn't an option at the moment
        pattern = re.compile(r'^Argument \(.+?\): (.+)$')

        type_mapping = {
            "int": "int",
            "string": "str",
            "boolean": "bool",
            "float": "float"
        }

        code = """
parser = ArgumentParser()        
        """
        body = ast.parse(code).body

        nodes = self._build_node_set()
        argument_nodes = [n for n in nodes if n.name.startswith("Argument ") and n.file is None]
        for arg in argument_nodes:
            m = pattern.match(arg.name)
            arg_name = m.group(1)
            tpl = "parser.add_argument('--%s', type=%s)" % (arg_name, type_mapping[arg.type])
            body.extend(ast.parse(tpl).body)

        return body



import json

from xircuits.compiler.parser import XircuitsFileParser
from xircuits.compiler.generator import CodeGenerator


def compile(input_file, output_file, component_python_paths=None):
    if component_python_paths is None:
        component_python_paths = {}

    parser = XircuitsFileParser()
    graph = parser.parse(input_file)
    generator = CodeGenerator(graph, component_python_paths)
    generator.generate(output_file)


from tqdm import tqdm
import os
from urllib import request, parse
from github import Github, GithubException
from typing import Optional
from .request_submodule import get_submodules
import subprocess
import re 

def request_folder(folder, repo_name="XpressAi/Xircuits", branch="master"):
    print("Downloading " + folder + " from " + repo_name + " branch " + branch)
    g = Github()
    
    try:
        repo = g.get_repo(repo_name)
        contents = repo.get_contents(folder, ref=branch)
    except GithubException as e:
        if e.status == 403:
            print("pyGithub API rate limit exceeded. If you're trying to fetch Xircuits components, you can use `xircuits-components`.")
        else:
            print(folder + " from " + repo_name + " branch " + branch + " does not exist!")
        return 
    except Exception as e:
        print("An error occurred: " + str(e))
        return

    if not os.path.exists(folder):
        os.mkdir(folder)
    else:
        print(folder + " already exists.")
    
    base_url = "https://raw.githubusercontent.com/" + repo_name + "/" + branch    
    urls = {}
    
    while len(contents)>0:
        file_content = contents.pop(0)
        if file_content.type=='dir':
            if not os.path.exists(file_content.path):
                os.mkdir(file_content.path)
            contents.extend(repo.get_contents(file_content.path, ref=branch))

        else:
            file_url = base_url + "/" + parse.quote(file_content.path)
            urls.update({file_url: file_content.path})

    submodules = get_submodules(repo, branch)

    for url in tqdm(urls):
        try:
            request.urlretrieve(url, urls[url])
        except:
            if urls[url] not in submodules:
                print("Unable to retrieve " + urls[url] + ". Skipping...")


def extract_library_details_from_url(github_url):
    """Extract organization and repository name from GitHub URL."""
    match = re.search(r'github.com/([^/]+)/xai-(.+)$', github_url)
    if not match:
        raise ValueError("Invalid GitHub URL format.")

    org_name = match.group(1)
    repo_name = match.group(2).replace('-', '_')
    return org_name, repo_name

def clone_repo(github_url, target_path):
    """Clone a repository from GitHub URL to the specified target path."""
    try:
        subprocess.run(["git", "clone", github_url, target_path], check=True)
    except subprocess.CalledProcessError:
        print(f"Error: Unable to clone {github_url} into {target_path}. The directory may already exist and is not empty.")
        return target_path

    return target_path

def clone_from_github_url(github_url: str) -> str:
    # Create a Github instance
    g = Github()

    org_name, repo_name = extract_library_details_from_url(github_url)
    target_path = f"xai_components/xai_{repo_name}"

    # Retrieve the repository
    try:
        repo = g.get_repo(f"{org_name}/xai-{repo_name}")

        # Check if repo exists, otherwise GithubException will be raised
        if repo:
            return clone_repo(github_url, target_path)
        else:
            raise ValueError(f"No repository found at {github_url}")

    except GithubException as e:
        raise RuntimeError(f"Error accessing the repository: {e}")


from xai_components.base import InArg, OutArg, InCompArg, Component, BaseComponent, xai_component

@xai_component
class SKLearnLoadDataset(Component):
    """
    Fetches a specified toy dataset from sklearn's dataset module.

    #### Reference:
    - [sklearn toy datasets](https://scikit-learn.org/stable/datasets/toy_dataset.html)

    ##### inPorts:
    - dataset_name: The name of the dataset to be loaded.

    ##### outPorts:
    - dataset: The loaded sklearn toy dataset.

    """
    dataset_name: InCompArg[str]
    dataset: OutArg[any]

    def execute(self, ctx) -> None:

        from sklearn import datasets

        # If the name is already prefixed with "load_", use it as is. Otherwise, add the prefix.
        name = self.dataset_name.value if self.dataset_name.value.startswith("load_") else f"load_{self.dataset_name.value}"
        
        try:
            load_func = getattr(datasets, name)
        except AttributeError:
            raise ValueError(f"No dataset named '{name}' found in sklearn.datasets")
        
        self.dataset.value = load_func()


@xai_component
class SKLearnTrainTestSplit(Component):
    """"
    Takes a sklearn dataset into train and test splits.

    #### Reference:
    - [sklearn.model_selection.train_test_split](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html)

    ##### inPorts:
    - dataset: The input sklearn dataset to be split.
    - train_split: The proportion of the dataset to include in the train split (default is 0.75).
    - random_state: The seed used by the random number generator (default is None).
    - shuffle: Whether or not to shuffle the data before splitting (default is True).
    - stratify: If not None, data is split in a stratified fashion, using this as the class labels (default is None).

    ##### outPorts:
    - X_train: The training data.
    - X_test: The testing data.
    - y_train: The target variable for the training data.
    - y_test: The target variable for the testing data.
    """

    dataset: InCompArg[any]
    train_split: InArg[float]
    random_state: InArg[int]
    shuffle: InArg[bool]
    stratify: InArg[any]
    X_train: OutArg[any] 
    X_test: OutArg[any] 
    y_train: OutArg[any] 
    y_test: OutArg[any] 

    def __init__(self):
        super().__init__()
        self.train_split.value = 0.75
        self.shuffle.value = True

    def execute(self, ctx) -> None:
        
        from sklearn.model_selection import train_test_split

        print(f"Split Parameters:\nTrain Split {self.train_split.value} \nShuffle: {self.shuffle.value} \nRandom State: {self.random_state.value}")
        self.X_train.value, self.X_test.value, self.y_train.value, self.y_test.value = train_test_split(self.dataset.value['data'], self.dataset.value['target'], 
                                    test_size=self.train_split.value, shuffle=self.shuffle.value, 
                                    random_state=self.random_state.value, stratify=self.stratify.value)
        
        print(f"Train data shape: {self.X_train.value.shape}, Train target shape: {self.y_train.value.shape}")
        print(f"Test data shape: {self.X_test.value.shape}, Test target shape: {self.y_test.value.shape}")


@xai_component
class CSVToSKLearnDataset(Component):
    """
    Transforms a CSV file into a format compatible with sklearn.datasets.

    This component reads a CSV file, selects specific columns to use as features and a target, converts it to a pandas DataFrame, and then transforms the DataFrame into a format compatible with sklearn.datasets. If the target column is categorical, it will be label encoded to numerical values.

    #### Reference:
    - [Pandas](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_csv.html)
    - [sklearn.datasets](https://scikit-learn.org/stable/datasets/toy_dataset.html)
    - [sklearn LabelEncoder](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html)

    ##### inPorts:
    - file_path: The path to the CSV file to be transformed.
    - feature_columns: The list of columns in the CSV file to use as data. If not specified, all columns except the target will be used.
    - target_column: The column in the CSV file to use as the target variable.
    - drop_na_rows: If set to True, rows with any NA/missing values will be dropped. Defaults to False.

    ##### outPorts:
    - dataset: The sklearn compatible dataset.
    """
    file_path: InArg[str]
    feature_columns: InArg[list]
    target_column: InArg[str]
    drop_na_rows: InArg[bool]
    dataset: OutArg[dict]

    def __init__(self):
        super().__init__()
        self.drop_na_rows.value = False
    
    def execute(self, ctx) -> None:
        import pandas as pd
        from sklearn.preprocessing import LabelEncoder

        df = pd.read_csv(self.file_path.value)

        # If feature_columns are not provided, use all columns except the target_column
        if not self.feature_columns.value:
            self.feature_columns.value = df.columns.drop(self.target_column.value).tolist()

        # Use only selected columns as features
        df = df[self.feature_columns.value + [self.target_column.value]]

        # If drop_na_rows is True, drop any rows with NA/missing values
        if self.drop_na_rows.value:
            df = df.dropna()

        target = df.pop(self.target_column.value)
        # If the target data type is object (likely strings), label encode it
        if target.dtype == 'object':
            le = LabelEncoder()
            target = le.fit_transform(target)

        data = df.values

        self.dataset.value = {
            'data': data,
            'target': target,
            'feature_names': df.columns.tolist(),
            'DESCR': f'Dataset loaded from {self.file_path.value}, target column is {self.target_column.value}'
        }
        
        print(f"Data shape: {data.shape}, Target shape: {target.shape}")




from argparse import Namespace
from typing import TypeVar, Generic, Tuple, NamedTuple, List

T = TypeVar('T')


class InArg(Generic[T]):
    value: T

    def __init__(self, value: T) -> None:
        self.value = value

    @classmethod
    def empty(cls):
        return InArg(None)


class OutArg(Generic[T]):
    value: T

    def __init__(self, value: T) -> None:
        self.value = value

    @classmethod
    def empty(cls):
        return OutArg(None)

class InCompArg(Generic[T]):
    value: T

    def __init__(self, value: T) -> None:
        self.value = value

    @classmethod
    def empty(cls):
        return InCompArg(None)


def xai_component(*args, **kwargs):
    # Passthrough element without any changes.
    # This is used for parser metadata only.
    if len(args) == 1 and callable(args[0]):
        # @xai_components form
        return args[0]
    else:
        # @xai_components(...) form
        def passthrough(f):
            return f
        return passthrough

class ExecutionContext:
    args: Namespace

    def __init__(self, args: Namespace):
        self.args = args


class BaseComponent:

    def __init__(self):
        all_ports = self.__annotations__
        for key, type_arg in all_ports.items():
            if isinstance(type_arg, InArg[any].__class__):
                setattr(self, key, InArg.empty())
            elif isinstance(type_arg, InCompArg[any].__class__):
                setattr(self, key, InCompArg.empty())
            elif isinstance(type_arg, OutArg[any].__class__):
                setattr(self, key, OutArg.empty())
            elif type_arg == str(self.__class__):
                setattr(self, key, None)

    @classmethod
    def set_execution_context(cls, context: ExecutionContext) -> None:
        cls.execution_context = context

    def execute(self, ctx) -> None:
        pass

    def do(self, ctx) -> Tuple[bool, 'BaseComponent']:
        pass


class Component(BaseComponent):
    next: BaseComponent

    def do(self, ctx) -> Tuple[bool, BaseComponent]:
        print(f"\nExecuting: {self.__class__.__name__}")
        self.execute(ctx)

        return self.next

    def debug_repr(self) -> str:
        return "<h1>Component</h1>"


class SubGraphExecutor:
    
    def __init__(self, component):
        self.comp = component
        
    def do(self, ctx):
        comp = self.comp
        
        while comp is not None:
            comp = comp.do(ctx)
        return None


def execute_graph(args: Namespace, start: BaseComponent, ctx) -> None:
    BaseComponent.set_execution_context(ExecutionContext(args))

    if 'debug' in args and args['debug']:
        import pdb
        pdb.set_trace()

        current_component = start
        next_component = current_component.do(ctx)
        while next_component:
            current_component = next_component
            next_component = current_component.do(ctx)
    else:
        next_component = start.do(ctx)
        while next_component:
            next_component = next_component.do(ctx)
            

class secret:

    def __init__(self, value):
        self.__value = value

    def get_value(self):
        return self.__value
    
    
class message(NamedTuple):
    role: str
    content: str

class chat(NamedTuple):
    messages: List[message]
    
class dynalist:
    def __init__(self, value):
        self.value = value

class dynatuple:
    def __init__(self, value):
        self.value = value

class dynadict:
    def __init__(self, value):
        self.value = value


from typing import Tuple, Dict
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import datasets, layers, models
import numpy as np
from xai_components.base import InArg, InCompArg, OutArg, Component, xai_component
import os
import sys
from pathlib import Path

@xai_component
class ReadKerasDataSet(Component):
    """Loads a Keras image dataset or creates a dataset from a directory.
    
    ### Reference:
    - Keras Model Applications: https://keras.io/api/datasets/

    ##### inPorts:
    - dataset_name: Loads a Keras image dataset given a valid string 
        (mnist / mnist fasion / cifar10 / cifar 100) OR 
        creates a dataset object when given a valid dataset directory path. 
        
        For the latter, the directory must have subdirectories and each 
        subdirectory name will be treated as its own class.

        for example, when given a Literal String `DATASET`, the structure must be:
        working_dir/
            |- DATASET 
                |- CLASS_1
                |- CLASS_2
                |- CLASS_3
        
    ##### outPorts:
    - dataset: a dataset tuple
    - class_dict: dict of classes if not using IMAGENET.
    """  

    dataset_name: InCompArg[str]
    dataset: OutArg[Tuple[np.array, np.array]]
    class_dict: OutArg[dict]

    def execute(self, ctx) -> None:

        if self.dataset_name.value == 'mnist':
            (x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()

            # Scale images to the [0, 1] range
            x_train = x_train.astype("float32") / 255
            x_test = x_test.astype("float32") / 255
            # Make sure images have shape (28, 28, 1)
            x_train = np.expand_dims(x_train, -1)
            x_test = np.expand_dims(x_test, -1)

            new_x = np.vstack((x_train, x_test))
            new_y = np.concatenate((y_train, y_test), axis=None)
            self.dataset.value = (new_x, new_y)

        elif self.dataset_name.value == 'fashion_mnist':
            (x_train, y_train), (x_test, y_test) = keras.datasets.fashion_mnist.load_data()

            # Scale images to the [0, 1] range
            x_train = x_train.astype("float32") / 255
            x_test = x_test.astype("float32") / 255
            # Make sure images have shape (28, 28, 1)
            x_train = np.expand_dims(x_train, -1)
            x_test = np.expand_dims(x_test, -1)

            new_x = np.vstack((x_train, x_test))
            new_y = np.concatenate((y_train, y_test), axis=None)
            self.dataset.value = (new_x, new_y)

        elif self.dataset_name.value == 'cifar10':
            (x_train, y_train), (x_test, y_test) = keras.datasets.cifar10.load_data()

            # Normalize pixel values to be between 0 and 1
            x_train, x_test = x_train / 255.0, x_test / 255.0

            new_x = np.vstack((x_train, x_test))
            new_y = np.concatenate((y_train, y_test), axis=None)

            self.dataset.value = (new_x, new_y)

        elif self.dataset_name.value == 'cifar100':
            (x_train, y_train), (x_test, y_test) = keras.datasets.cifar10.load_data()

            # Normalize pixel values to be between 0 and 1
            x_train, x_test = x_train / 255.0, x_test / 255.0

            new_x = np.vstack((x_train, x_test))
            new_y = np.concatenate((y_train, y_test), axis=None)

            self.dataset.value = (new_x, new_y)


        elif self.dataset_name.value:
            try:
                import cv2
                from tqdm import tqdm

                BASE_FOLDER = self.dataset_name.value
                folders = [os.path.join(BASE_FOLDER, folder) for folder in os.listdir(BASE_FOLDER)]
                
                print(f"Detecting {len(folders)} classes in {BASE_FOLDER}.")
                # lists to store data
                data = []
                label = []
                for folder in tqdm(folders):
                    for file in os.listdir(folder):

                        file = os.path.join(folder, file)

                        try:
                            img = cv2.imread(file)
                            img = cv2.resize(img, (256, 256))
                            data.append(img)
                            label.append(folder)

                        except: 
                            print(f'Error reading file: {os.path.abspath(file)}. Skipping...')                        

                new_x = np.asarray(data)

                # Import label encoder
                from sklearn import preprocessing
                label_encoder = preprocessing.LabelEncoder()
                new_y = label_encoder.fit_transform(label)

                print(f"x_shape = {new_x.shape}, y_shape = {new_y.shape}")

                self.dataset.value = (new_x, new_y)

            except Exception as e: 
                print(e)

        else:
            print("Dataset was not found!")


@xai_component
class FlattenImageData(Component):

    """Takes a 2D dataset tuple from the ReadDataSet component  
    that contains tuple and flattens it to 1D. 

    ##### inPorts:
    - dataset: 2D dataset tuple from the ReadDataSet component.

    ##### outPorts:
    - resized_dataset: 1D tuple dataset object.
    """  

    dataset: InCompArg[Tuple[np.array, np.array]]
    resized_dataset: OutArg[Tuple[np.array, np.array]]

    def execute(self, ctx) -> None:

        x = self.dataset.value[0]
        x = x.reshape(x.shape[0], -1)

        self.resized_dataset.value = (x, self.dataset.value[1])
        print(f"resized_dataset = {np.shape(self.resized_dataset.value)}")


@xai_component
class TrainTestSplit(Component):
    """Takes a dataset tuple and splits it into train test tuples.
    
    ### Reference: 
    - [Scikit-learn Train Test Split](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html)

    ##### inPorts:
    - dataset: dataset in the form of np array tuples. `Tuple[np.array, np.array]`
    - train_split: float ratio of the train split. Default `0.75`.
    - random_state: seed for random state. Default `None`.
    - shuffle: Enable dataset shuffle with True / False. Default `True`.
    - stratify: Data is split in a stratified fashion, using this as the class labels. 
        Default `None`.

    ##### outPorts:
    - train: tuple that contains the train split of the dataset.
    - test:tuple that contains the test split of the dataset.
    - resized_dataset: 1D tuple object.
    """

    dataset: InCompArg[Tuple[np.array, np.array]]
    train_split: InArg[float]
    random_state: InArg[int]
    shuffle: InArg[bool]
    stratify: InArg[any]
    train: OutArg[Tuple[np.array, np.array]] 
    test: OutArg[Tuple[np.array, np.array]] 

    def execute(self, ctx) -> None:
        
        from sklearn.model_selection import train_test_split

        train_split = self.train_split.value if self.train_split.value else 0.75
        shuffle = self.shuffle.value if self.shuffle.value else True
        random_state = self.random_state.value if self.random_state.value else None
        stratify = self.stratify.value if self.stratify.value else None

        print(f"Split Parameters:\nTrain Split {train_split} \nShuffle: {shuffle} \nRandom State: {random_state}")
        splits = train_test_split(self.dataset.value[0], self.dataset.value[1], 
                                    test_size=train_split, shuffle=shuffle, 
                                    random_state=random_state, stratify=stratify)
        
        train_x = splits[0]
        test_x = splits[1]
        train_y = splits[2]
        test_y = splits[3]

        train = train_x, keras.utils.to_categorical(train_y, int(test_y.max()) + 1)
        test = test_x, keras.utils.to_categorical(test_y, int(test_y.max()) + 1)

        self.train.value = train
        self.test.value = test
@xai_component
class KerasCreate1DInputModel(Component):
    """Takes a 1D dataset tuple and creates a 1D Keras model.

    ##### inPorts:
    - training_data: dataset tuple which contains 1D numpy array.
    
    ##### outPorts:
    - model: keras model.
    """

    training_data: InCompArg[Tuple[np.array, np.array]]
    model: OutArg[keras.Sequential]

    def execute(self, ctx) -> None:
        x_shape = self.training_data.value[0].shape
        y_shape = self.training_data.value[1].shape

        model = keras.Sequential([
            keras.layers.Dense(512, activation='relu', input_shape=(x_shape[1],)),
           keras.layers.Dropout(rate=0.5),
            keras.layers.Dense(y_shape[1], activation='softmax')
        ])

        model.compile(
            loss='categorical_crossentropy',
            optimizer='adam',
            metrics=['accuracy']
        )

        self.model.value = model

@xai_component
class KerasCreate2DInputModel(Component):
    """Takes a 2D dataset tuple and creates a 2D Keras model.

    ##### inPorts:
    - training_data: dataset tuple which contains 2D numpy array.
    
    ##### outPorts:
    - model: keras model.
    - model_config: keras model config dict. 
        Contains 'lr', 'optimizer_name' and 'loss'.
    """
    training_data: InCompArg[Tuple[np.array, np.array]]

    model: OutArg[keras.Sequential]
    model_config: OutArg[dict]

    def execute(self, ctx) -> None:

        x_shape = self.training_data.value[0].shape[1:]
        y_shape = self.training_data.value[1].shape[1]
        print(f"{x_shape=}")
        print(f"{y_shape=}")

        model = keras.Sequential(
            [
                keras.Input(shape=x_shape),
                layers.Conv2D(32, kernel_size=(3, 3), activation="relu"),
                layers.MaxPooling2D(pool_size=(2, 2)),
                layers.Conv2D(64, kernel_size=(3, 3), activation="relu"),
                layers.MaxPooling2D(pool_size=(2, 2)),
                layers.Flatten(),
                layers.Dropout(0.5),
                layers.Dense(y_shape, activation="softmax"),
            ]
        )

        model.compile(
            loss='categorical_crossentropy',
            optimizer='adam',
            metrics=['accuracy']
        )
                
        model_config = {
            'lr': model.optimizer.lr.numpy().item(),
            'optimizer_name': model.optimizer.name,
            'loss': model.loss,
        }

        self.model.value = model
        self.model_config.value = model_config


@xai_component
class KerasTrainImageClassifier(Component):
    """Trains a Keras model for image classification.

    ##### inPorts:
    - model: a Keras model object.
    - training_data: a dataset tuple with (X (data), Y (label)).
    - training_epochs: number of training epochs. Default `1`.

    ##### outPorts:
    - trained_model: trained Keras model config.
    - training_metrics: dict which contains results of training.
    """
    
    model: InCompArg[keras.Sequential]
    training_data: InCompArg[Tuple[np.array, np.array]] 
    training_epochs: InArg[int]

    trained_model: OutArg[keras.Sequential]
    training_metrics: OutArg[dict]

    def execute(self, ctx) -> None:

        model = self.model.value
        epoch = self.training_epochs.value if self.training_epochs.value else 1

        train = model.fit(
            self.training_data.value[0],
            self.training_data.value[1],
            batch_size=32,
            epochs=epoch
        )

        # Set training metrics
        training_metrics = {}
        for key in train.history.keys():
            training_metrics[key] = {}
            [training_metrics[key].update({i + 1: v}) for i, v in enumerate(train.history[key])]

        self.trained_model.value = model
        self.training_metrics.value = training_metrics

@xai_component
class KerasEvaluateAccuracy(Component):
    """Evaluates a Keras model against a dataset

    ##### inPorts:
    - model: a Keras model object.
    - eval_dataset: a dataset tuple with (X (data), Y (label)).

    ##### outPorts:
    - metrics: dict which contains results of evaluation.

    """
    model: InCompArg[keras.Sequential]
    eval_dataset: InCompArg[Tuple[np.array, np.array]]

    metrics: OutArg[Dict[str, str]]

    def execute(self, ctx) -> None:
        (loss, acc) = self.model.value.evaluate(self.eval_dataset.value[0], self.eval_dataset.value[1], verbose=0)
        metrics = {
            'loss': str(loss),
            'accuracy': str(acc)
        }
        print(metrics)

        self.metrics.value = metrics


@xai_component
class ShouldStop(Component):
    """Checks whether model evaluation has reached targeted accuracy.

    ##### inPorts:
    - target_accuracy: the targeted accuracy in floats.
    - max_retries: the number of attempted tries. Default `1`.
    - metrics: dict that contains results of evaluation.

    ##### outPorts:
    - should_retrain: True if targeted accuracy not reached.
    
    """
    target_accuracy: InCompArg[float]
    metrics: InCompArg[Dict[str, str]]
    max_retries: InArg[int]

    should_retrain: OutArg[bool]

    def __init__(self):
        super().__init__()
        self.should_retrain.value = True
        self.retries = 0

    def execute(self, ctx) -> None:
        self.retries += 1
        max_retries = self.max_retries.value if self.max_retries.value else 1

        if self.retries < max_retries:
            the_accuracy = float(self.metrics.value['accuracy'])
            print('Eval accuracy:' + str(the_accuracy))

            if the_accuracy < self.target_accuracy.value:
                print('Will retrain')
                self.should_retrain.value = True
            else:
                print('Target accuracy achieved')
                self.should_retrain.value = False
        else:
            print('Unable to achieve target accuracy.  Giving up.')
            self.should_retrain.value = False
@xai_component
class SaveKerasModel(Component):
    """Saves current Keras model.

    ##### inPorts:
    - model: a Keras model.
    - model_name: name to save the Keras model. Default is the .xircuits file name.

    ##### outPorts:
    - model_h5_path: path of the generated .h5 model.
    
    """
    model: InCompArg[any]
    model_name: InArg[str]
    model_h5_path: OutArg[str]

    def execute(self, ctx) -> None:
        model = self.model.value
        model_name = self.model_name.value if self.model_name.value else os.path.splitext(sys.argv[0])[0] + ".h5"
        model.save(model_name)
        print(f"Saving Keras h5 model at: {model_name}")
        self.model_h5_path.value = model_name


from xai_components.base import InArg, InCompArg, OutArg, Component, xai_component

import os
import sys

import torch
from torch import nn


@xai_component
class TorchModel(Component):
    """Creates a custom Torch Model config.

    ##### outPorts:
    - model_in: List of layers to make into Sequential Model
    - model_config: resulting model.
    - loss_fn: nn.CrossEntropyLoss()
    - optimizer: torch.optim.SGD(model.parameters(), lr=1e-3)
    """

    model_in: InArg[list]
    loss_in: InArg[str]
    learning_rate = InArg[float]
    optimizer_in: InArg[str]
    should_flatten: InArg[bool]
    model_config: OutArg[nn.Module]
    loss_fn: OutArg[any]
    optimizer: OutArg[any]

    def __init__(self):
        super().__init__()
        self.learning_rate.value = 1e-3
        self.should_flatten.value = False

    def execute(self,ctx) -> None:
        should_flatten = self.should_flatten.value
        stack = self.model_in.value
        
        # Define model
        class NeuralNetwork(nn.Module):
            def __init__(self):
                super(NeuralNetwork, self).__init__()
                
                if should_flatten:
                    self.flatten = nn.Flatten()
                self.stack = nn.Sequential(*stack)

            def forward(self, x):
                if should_flatten:
                    x = self.flatten(x)
                    
                logits = self.stack(x)
                return logits

        # Get cpu or gpu device for training.
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {device} device")
        model = NeuralNetwork().to(device)
        print(model)

        loss_fn = nn.MSELoss()
        if self.loss_in.value == 'CrossEntropyLoss':
            loss_fn = nn.CrossEntropyLoss()
        elif self.loss_in.value == 'L1Loss':
            loss_fn = nn.L1Loss()
        elif self.loss_in.value == 'CTCLoss':
            loss_fn = nn.CTCLoss()
        else:
            loss_fn = eval(self.loss_in.value)

        
        optimizer = torch.optim.SGD(model.parameters(), lr=self.learning_rate.value)
        if self.optimizer_in.value == 'Adam':
            optimizer = torch.optim.Adam(model.parameters(), lr=self.learning_rate.value)
        elif self.optimizer_in.value == 'RMSprop':
            optimizer = torch.optim.RMSprop(model.parameters(), lr=self.learning_rate.value)
        else:
            optimizer = eval(self.optimizer_in.value)

        self.model_config.value = model
        self.loss_fn.value = loss_fn
        self.optimizer.value = optimizer


@xai_component
class TorchAddLinearLayer(Component):
    """Adds a LinearLayer to a sequential model."""

    model_in: InArg[list]
    in_features: InArg[int]
    out_features: InArg[int]
    bias: InArg[bool]
    model_out: OutArg[list]

    def execute(self,ctx) -> None:
        bias = True if self.bias.value is None else False
        in_size = self.in_features.value
        out_size = self.out_features.value
        
        if self.model_in.value is None:
            self.model_out.value = [nn.Linear(in_size, out_size, bias)]
        else:
            self.model_out.value = self.model_in.value + [nn.Linear(in_size, out_size, bias)]


@xai_component
class TorchAddConv1DLayer(Component):
    """Adds a Conv1DLayer to a sequential model."""

    model_in: InArg[list]

    in_channels: InCompArg[int]
    out_channels: InCompArg[int]
    kernel_size: InCompArg[any] #int or tuple
    stride: InArg[any] #int or tuple
    padding: InArg[any] #int, tuple or str
    dilation: InArg[any] #int or tuple
    groups: InArg[int]
    bias: InArg[bool]
    padding_mode: InArg[str]

    model_out: OutArg[list]


    def __init__(self):
        super().__init__()
        self.stride.value = 1
        self.padding.value = 0
        self.dilation.value = 1
        self.groups.value = 1
        self.bias.value = True
        self.padding_mode.value = 'zeros'
        

    def execute(self,ctx) -> None:

        in_channels = self.in_channels.value 
        out_channels = self.out_channels.value 
        kernel_size = self.kernel_size.value 
        stride = self.stride.value 
        padding = self.padding.value 
        dilation = self.dilation.value 
        groups = self.groups.value 
        bias = self.bias.value 
        padding_mode = self.padding_mode.value 
        
        if self.model_in.value is None:
            self.model_out.value = [nn.Conv1d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode)]
        else:
            self.model_out.value = self.model_in.value + [nn.Conv1d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode)]


@xai_component
class TorchAddConv2DLayer(Component):
    """Adds a Conv2DLayer to a sequential model."""

    model_in: InArg[list]

    in_channels: InCompArg[int]
    out_channels: InCompArg[int]
    kernel_size: InCompArg[any] #int or tuple
    stride: InArg[any] #int or tuple
    padding: InArg[any] #int, tuple or str
    dilation: InArg[any] #int or tuple
    groups: InArg[int]
    bias: InArg[bool]
    padding_mode: InArg[str]

    model_out: OutArg[list]


    def __init__(self):
        super().__init__()
        self.stride.value = 1
        self.padding.value = 0
        self.dilation.value = 1
        self.groups.value = 1
        self.bias.value = True
        self.padding_mode.value = 'zeros'

    def execute(self,ctx) -> None:

        in_channels = self.in_channels.value 
        out_channels = self.out_channels.value 
        kernel_size = self.kernel_size.value 
        stride = self.stride.value 
        padding = self.padding.value 
        dilation = self.dilation.value 
        groups = self.groups.value 
        bias = self.bias.value 
        padding_mode = self.padding_mode.value 
        
        if self.model_in.value is None:
            self.model_out.value = [nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode)]
        else:
            self.model_out.value = self.model_in.value + [nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode)]

@xai_component
class TorchAddTransformerEncoderLayer(Component):
    """Adds a TransformerEncoderLayer to a sequential model."""

    model_in: InArg[list]

    d_model : InCompArg[int]
    nhead : InCompArg[int]
    dim_feedforward : InArg[int]
    dropout : InArg[float]
    activation : InArg[any] #Union[str, Callable[[Tensor], Tensor]]
    layer_norm_eps : InArg[float]
    batch_first : InArg[bool]
    norm_first : InArg[bool]

    model_out: OutArg[list]


    def __init__(self):
        super().__init__()

        self.dim_feedforward.value = 2048
        self.dropout.value = 0.1
        self.activation.value = 'relu'
        self.layer_norm_eps.value = 1e-05
        self.batch_first.value = False
        self.norm_first.value = False

    def execute(self,ctx) -> None:

        d_model = self.d_model.value
        nhead = self.nhead.value
        dim_feedforward = self.dim_feedforward.value
        dropout = self.dropout.value
        activation = self.activation.value
        layer_norm_eps = self.layer_norm_eps.value
        batch_first = self.batch_first.value
        norm_first = self.norm_first.value
        
        if self.model_in.value is None:
            self.model_out.value = [nn.TransformerEncoderLayer(d_model, nhead, dim_feedforward, dropout, activation, layer_norm_eps, batch_first, norm_first)]
        else:
            self.model_out.value = self.model_in.value + [nn.TransformerEncoderLayer(d_model, nhead, dim_feedforward, dropout, activation, layer_norm_eps, batch_first, norm_first)]

@xai_component
class TorchAddTransformerDecoderLayer(Component):
    """Adds a TransformerDecoderLayer to a sequential model."""

    model_in: InArg[list]

    d_model : InCompArg[int]
    nhead : InCompArg[int]
    dim_feedforward : InArg[int]
    dropout : InArg[float]
    activation : InArg[any] #Union[str, Callable[[Tensor], Tensor]]
    layer_norm_eps : InArg[float]
    batch_first : InArg[bool]
    norm_first : InArg[bool]

    model_out: OutArg[list]


    def __init__(self):
        super().__init__()
        self.dim_feedforward.value = 2048
        self.dropout.value = 0.1
        self.activation.value = 'relu'
        self.layer_norm_eps.value = 1e-05
        self.batch_first.value = False
        self.norm_first.value = False

    def execute(self,ctx) -> None:

        d_model = self.d_model.value
        nhead = self.nhead.value
        dim_feedforward = self.dim_feedforward.value
        dropout = self.dropout.value
        activation = self.activation.value
        layer_norm_eps = self.layer_norm_eps.value
        batch_first = self.batch_first.value
        norm_first = self.norm_first.value
        
        if self.model_in.value is None:
            self.model_out.value = [nn.TransformerDecoderLayer(d_model, nhead, dim_feedforward, dropout, activation, layer_norm_eps, batch_first, norm_first)]
        else:
            self.model_out.value = self.model_in.value + [nn.TransformerDecoderLayer(d_model, nhead, dim_feedforward, dropout, activation, layer_norm_eps, batch_first, norm_first)]


@xai_component
class TorchLSTM(Component):
    """Adds a LSTM to a sequential model."""

    model_in: InArg[list]

    input_size: InCompArg[int]
    hidden_size: InCompArg[int]
    num_layers: InArg[int]
    bias: InArg[bool]
    batch_first: InArg[bool]
    dropout: InArg[float]
    bidirectional: InArg[bool]
    proj_size: InArg[int]

    model_out: OutArg[list]


    def __init__(self):
        super().__init__()

        self.num_layers.value = 1
        self.bias.value = True
        self.batch_first.value = False
        self.dropout.value = 0
        self.bidirectional.value = False
        self.proj_size.value = 0

    def execute(self,ctx) -> None:

        input_size = self.input_size.value
        hidden_size = self.hidden_size.value
        num_layers = self.num_layers.value
        bias = self.bias.value
        batch_first = self.batch_first.value
        dropout = self.dropout.value
        bidirectional = self.bidirectional.value
        proj_size = self.proj_size.value

        if self.model_in.value is None:
            self.model_out.value = [nn.LSTM(input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional, proj_size)]
        else:
            self.model_out.value = self.model_in.value + [nn.LSTM(input_size, hidden_size, num_layers, bias, batch_first, dropout, bidirectional, proj_size)]

@xai_component
class TorchAddReluLayer(Component):
    """Adds a Relu activation to a sequential model."""
    
    model_in: InArg[list]
    model_out: OutArg[list]

    def execute(self, ctx) -> None:
        if self.model_in.value is None:
            self.model_out.value = [nn.ReLU()]
        else:
            self.model_out.value = self.model_in.value + [nn.ReLU()]

@xai_component
class TorchAddDropoutLayer(Component):
    """Adds a Dropout to a sequential model."""
    
    model_in: InArg[list]
    prob_zero: InArg[float]
    model_out: OutArg[list]

    
    def execute(self, ctx) -> None:
        prob = 0.5
        if self.prob_zero.value is not None:
            prob = self.prob_zero.value
            
        if self.model_in.value is None:
            self.model_out.value = [nn.Dropout(prob)]
        else:
            self.model_out.value = self.model_in.value + [nn.Dropout(prob)]


from jupyter_server.utils import url_path_join

from .compile_xircuits import CompileXircuitsFileRouteHandler
from .components import ComponentsRouteHandler
from .config import RunConfigRouteHandler, SplitModeConfigHandler
from .debugger import DebuggerRouteHandler
from .spark_submit import SparkSubmitRouteHandler


def setup_handlers(web_app, url_path):
    host_pattern = ".*$"
    base_url = web_app.settings["base_url"]

    # Prepend the base_url so that it works in a JupyterHub setting
    web_app.add_handlers(host_pattern, [
        (
            url_path_join(base_url, url_path, "debug/enable"),
            DebuggerRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "config/run"),
            RunConfigRouteHandler
        ),
        (
        url_path_join(base_url, url_path, "config/split_mode"),
        SplitModeConfigHandler
        ),
        (
            url_path_join(base_url, url_path, "components/"),
            ComponentsRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "spark/submit"),
            SparkSubmitRouteHandler
        ),
        (
            url_path_join(base_url, url_path, "file/compile"),
            CompileXircuitsFileRouteHandler
        )

    ])